# Copyright 2019-2022 DADoES, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License in the root directory in the "LICENSE" file or at:
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import sys
import importlib
import logging
import anatools.lib.context as ctx

logger = logging.getLogger(__name__)

def create_node(name, alias):
    """Factory method for node class """

    try:
        module = ctx.channel.classes[alias]["module"]
    except KeyError:
        logger.critical("No module defined for node '%s' in channel definition file", alias)
        raise
    try:
        class_name = ctx.channel.classes[alias]["class"]
    except KeyError:
        logger.critical("No class defined for node '%s' in channel definition file", alias)
        raise

    # import the class
    try:
        node_class = getattr(importlib.import_module(module), class_name)
    except ModuleNotFoundError:
        logger.critical("Can't import node '%s' - module not found error", alias)
        raise
    except AttributeError:
        logger.critical("Can't import node '%s' - attribute error", alias)
        raise

    # instantiate the node
    node = node_class(name, alias)

    return node

# base class for all nodes
class Node:
    """Base class for all nodes"""
    def __init__(self, name, alias):

        logger.info("Creating node '%s', class = '%s'", name, self.__class__.__name__)

        self.name = name
        self.alias = alias
        self.properties = {}  # deprecated
        self.inputs = {}
        self.inlinks = {}
        self.outlinks = {}
        self.schema = ctx.channel.schemas[alias]
        self.input_types = ["values", "links"]
        # self.version =

    def configure(self, config):
        """Configure node"""

        logger.info("Configuring node '%s', class = '%s'", self.name, self.__class__.__name__)

        #
        # This function configures inputs and inlinks as follows:
        #
        #   self.inputs = {
        #        "inport1": [val1, val2, val3,...],
        #        ...
        #   }
        #
        #   self.inlinks: {
        #       "inport1": [(src_node1, src_port1), ...],
        #       ...
        #   }
        #
        # The interpreter will configure outlinks as follows:
        #
        #   self.outlinks: {
        #       "outport1": [(dst_node1, dst_port1), ...],
        #   }

        logger.debug("SCHEMA: %s", self.schema)

        # get set of all schema inputs and their defaults
        schema_input_set = set()
        input_defaults = {}
        if self.schema["inputs"] is not None:
            for port_dict in self.schema["inputs"]:
                schema_input_set.add(port_dict["name"])
                if "default" in port_dict:
                    input_defaults[port_dict["name"]] = port_dict.get("default")

        for key, value in config.get("values", {}).items():
            if key not in schema_input_set:
                logger.error("Input port '%s' is not defined in schema for class '%s'", key, type(self).__name__)
                sys.exit(1)

            if key not in self.inputs:
                self.inputs[key] = []
            self.inputs[key].append(value)

        for key, value in config.get("links", {}).items():
            # {nodeName: [{sourceNode: <src_node>, outputPort: <out>}]}
            if key not in self.inlinks:
                self.inlinks[key] = []
            for link in value:
                src_node = link['sourceNode']
                src_port = link['outputPort']
                if src_node == self.name:
                    logger.error("Node '%s' class '%s' port '%s' is connected to the same node",
                                  self.name, self.__class__.__name__, src_port)
                    sys.exit(1)

                self.inlinks[key].append((src_node, src_port))

            logger.debug("INPUTS: %s", self.inputs)

        # if a port doesn't have an input or inlink then use the default
        check_for_defaults = list(schema_input_set - (set(self.inputs.keys()) | set(self.inlinks.keys())))
        for port in check_for_defaults:
            if port in input_defaults.keys():
                self.inputs[port] = [input_defaults[port]]

        # make sure all inputs defined in the schema have either an input or inlink
        unresolveable_ports = list(schema_input_set - (set(self.inputs.keys()) | set(self.inlinks.keys())))
        if len(unresolveable_ports) > 0:
            errorString = "The following ports for node '{}' class '{}' cannot be resolved: ".format(
                self.name, self.__class__.__name__)
            errorString += ', '.join(unresolveable_ports)
            logger.error(errorString)
            sys.exit(1)

    def configure_v0(self, config):
        """Configure node"""

        logger.info("Configuring node '%s', class = '%s'", self.name, self.__class__.__name__)

        #
        # This function configures inputs and inlinks as follows:
        #
        #   self.inputs = {
        #        "inport1": [val1, val2, val3,...],
        #        ...
        #   }
        #
        #   self.inlinks: {
        #       "inport1": [(src_node1, src_port1), ...],
        #       ...
        #   }
        #
        # The interpreter will configure outlinks as follows:
        #
        #   self.outlinks: {
        #       "outport1": [(dst_node1, dst_port1), ...],
        #   }

        logger.debug("SCHEMA: %s", self.schema)

        # get set of all schema inputs and their defaults
        schema_input_set = set()
        input_defaults = {}
        if self.schema["inputs"] is not None:
            for port_dict in self.schema["inputs"]:
                schema_input_set.add(port_dict["name"])
                if "default" in port_dict:
                    input_defaults[port_dict["name"]] = port_dict.get("default")

        if config.get("inputs") is not None:
            for key, value in config["inputs"].items():
                if key not in schema_input_set:
                    logger.error("Input port '%s' is not defined in schema for class '%s'", key, type(self).__name__)
                    sys.exit(1)
                # "in1": {...}
                if isinstance(value, dict):
                    key1, value1 = next(iter(value.items()))
                    # "in1": {"$link": ["src_node1", "src_port1"]}
                    if key1 == "$link" and isinstance(value1, list) and len(value1) == 2:
                        # inlinks["in1"][0] = ("src_node1", "src_port1")
                        src_node = value1[0]
                        src_port = value1[1]
                        if src_node == self.name:
                            logger.error("Node '%s' class '%s' port '%s' is connected to the same node",
                                          self.name, self.__class__.__name__, src_port)
                            sys.exit(1)
                        if key not in self.inlinks:
                            self.inlinks[key] = []
                        self.inlinks[key].append((src_node, src_port))
                    # "in1": {"$list": [elem1, elem2, ...]}
                    elif key1 == "$list" and isinstance(value1, list):
                        for elem in value1:
                            # "in1": {"$list": [... {"key2": "value2"}, ...]}
                            if isinstance(elem, dict):
                                key2, value2 = next(iter(elem.items()))
                                # "in1": {"$list": [... {"$link": ["src_node1", "src_port1"]}, ...]}
                                if key2 == "$link" and isinstance(value2, list) and len(value2) == 2:
                                    # inlinks["in1"][n] = ("src_node1", "src_port1")
                                    src_node = value2[0]
                                    src_port = value2[1]
                                    if src_node == self.name:
                                        logger.error("Node '%s' class '%s' port '%s' is connected to the same node",
                                                    self.name, self.__class__.__name__, src_port)
                                        sys.exit(1)
                                    if key not in self.inlinks:
                                        self.inlinks[key] = []
                                    self.inlinks[key].append((src_node, src_port))
                                else:
                                    if key not in self.inputs:
                                        self.inputs[key] = []
                                    self.inputs[key].append(value1)
                            else:
                                if key not in self.inputs:
                                    self.inputs[key] = []
                                self.inputs[key].append(elem)
                    else:
                        if key not in self.inputs:
                            self.inputs[key] = []
                        self.inputs[key].append(value)
                else:
                    if key not in self.inputs:
                        self.inputs[key] = []
                    self.inputs[key].append(value)
            logger.debug("INPUTS: %s", self.inputs)

        # if a port doesn't have an input or inlink then use the default
        check_for_defaults = list(schema_input_set - (set(self.inputs.keys()) | set(self.inlinks.keys())))
        for port in check_for_defaults:
            if port in input_defaults.keys():
                self.inputs[port] = [input_defaults[port]]

        # make sure all inputs defined in the schema have either an input or inlink
        unresolveable_ports = list(schema_input_set - (set(self.inputs.keys()) | set(self.inlinks.keys())))
        if len(unresolveable_ports) > 0:
            errorString = "The following ports for node '{}' class '{}' cannot be resolved: ".format(
                self.name, self.__class__.__name__)
            errorString += ', '.join(unresolveable_ports)
            logger.error(errorString)
            sys.exit(1)

    def exec(self):
        """Execute Node"""
        return {}


# Copyright 2019-2022 DADoES, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License in the root directory in the "LICENSE" file or at:
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
import numpy as np
import anatools.lib.context as ctx
from anatools.lib.node import Node

logger = logging.getLogger(__name__)


class SweepArange(Node):
    """Do a parameter sweep across evenly spaced values within a half-open interval [start,stop)"""

    def exec(self):
        """Execute node"""
        logger.info("Executing {}".format(self.name))

        start = float(self.inputs["start"][0])
        stop = float(self.inputs["stop"][0])
        step = float(self.inputs["step"][0])

        values = np.arange(start, stop, step)
        value = values[ctx.interp_num % len(values)]

        logger.debug("Arange value = %s", value)

        return {"value": value}



"""
Staged Graphs Functions
"""

def get_staged_graphs(self, graphId=None, name=None, email=None, workspaceId=None):
    """Queries the workspace staged graphs based off provided parameters. Checks on graphId, name, or owner in this respective order within the specified workspace.
    If only workspace ID is provided, this will return all the staged graphs in a workspace. 
    
    Parameters
    ----------
    graphid : str
        Staged GraphID to filter on. Optional.
    name : str
        Name of the staged graph to filter on. Optional.
    email: str
        Owner of staged graphs to filter on. Optional.
    workspaceId : str    
        Workspace ID to filter on. If none is provided, the default workspace will get used. 
    
    Returns
    -------
    list[dict]
        A list of staged graphs based off provided query parameters if any parameters match.
    """
    if self.check_logout(): return
    if graphId is None: graphId = ''
    if name is None: name = ''
    if email is None: email = ''
    if workspaceId is None: workspaceId = self.workspace
    return self.ana_api.getGraphs(workspaceId=workspaceId, graphId=graphId, name=name, email=email)


def create_staged_graph(self, name, channelId, graph, description=None, workspaceId=None):
    """Generates a new staged graph based off provided parameters. Must provide valid json string to create a new staged graph.
    
    Parameters
    ----------
    name : str
        Name for the  that will get generated.
    channelId: str
        Id of channel to generate the staged graph with.
    graph: str
        The graph as a dictionary or `JSON` string. While `YAML` files are used in channel development, the Platform SDK and API only support `JSON`. 
        Ensure that the `YAML` file is valid in order for the `yaml.safe_load` to convert `YAML` to a dictionary for you. Otherwise,  provide a graph in `JSON` format.
    description: str
        Description of staged graph. Optional.
    workspaceId : str    
        Workspace ID create the staged graph in. If none is provided, the default workspace will get used. 
    
    Returns
    -------
    str
        The staged graph id if it was created sucessfully or an error message.
    """
    if self.check_logout(): return
    if name is None: raise ValueError('Name input must be defined.')
    if graph is None: raise ValueError('Graph input must be defined.')
    if channelId is None: raise ValueError('ChannelId input must be defined.')
    if workspaceId is None: workspaceId = self.workspace
    if type(graph) is dict:
        import json
        graph = json.dumps(graph)
    return self.ana_api.createGraph(workspaceId=workspaceId, channelId=channelId, graph=graph, name=name, description=description)


def edit_staged_graph(self, graphId, description=None, name=None, workspaceId=None):
    """Update staged graph description and name. 
    
    Parameters
    ----------
    graphId : str
        Staged Graph id to update.
    description: str
        New description to update.
    name: str
        New name to update.
    workspaceId : str    
        Workspace ID of the staged graph's workspace. If none is provided, the current workspace will get used. 
    
    Returns
    -------
    str
        A success or error message based on staged graph's update.
    """
    if self.check_logout(): return
    if graphId is None: return
    if name is None and description is None: return
    if workspaceId is None: workspaceId = self.workspace
    return self.ana_api.editGraph(workspaceId=workspaceId, graphId=graphId, name=name, description=description)


def delete_staged_graph(self, graphId, workspaceId=None):
    """Delete a staged graph in a workspace.
    
    Parameters
    ----------
    graphId : str
        Staged Graph id to delete.
    workspaceId : str    
        Workspace ID of the staged graph's workspace. If none is provided, the current workspace will get used. 
    
    Returns
    -------
    str
        A success or error message based on staged graph's delete.
    """
    if self.check_logout(): return
    if graphId is None: return
    if workspaceId is None: workspaceId = self.workspace
    return self.ana_api.deleteGraph(workspaceId=workspaceId, graphId=graphId)
    

def download_staged_graph(self, graphId, workspaceId=None):
    """Download a staged graph.
    
    Parameters
    ----------
    graphId : str
        Graph ID of the staged graph to download.
    workspaceId : str    
        Workspace ID of the staged graph's workspace. If none is provided, the default workspace will get used. 
    
    Returns
    -------
    str
        A download URL that can be used in the browser or a failure message.
    """
    if self.check_logout(): return
    if graphId is None: return
    if workspaceId is None: workspaceId = self.workspace
    return self.ana_api.downloadGraph(workspaceId=workspaceId, graphId=graphId)


def get_default_graph(self, channelId):
    """Gets the default graph for a channel.
    
    Parameters
    ----------
    channelId:
        Id of channel to get the default graph for.
   
    Returns
    -------
    json
        json data representing the graph.
    """
    if self.check_logout(): return
    return self.ana_api.getDefaultGraph(channelId=channelId)


def set_default_graph(self, channelId, workspaceId, graphId=None, stagedGraphId=None):
    """Sets the default graph for a channel. User must be in the organization that manages the channel.
    
    Parameters
    ----------
    channel : str
        The name of the channel to update the default graph.
    workspaceId : str
        The ID of the Workspace that the graph is in.
    graphId: str
        The ID of the graph that you want to be the default for the channel. Optional.
    stagedGraphId: str
        The ID of the staged graph that you want to be the default for the channel. Optional.

    Returns
    -------
    str
        Status
    """
    if self.check_logout(): return
    if workspaceId is None: workspaceId = self.workspace
    if graphId is None and stagedGraphId is None: raise Exception('GraphID or StagedGraphId must be specified.')
    return self.ana_api.setChannelGraph(channelId=channelId, workspaceId=workspaceId, graphId=graphId ,stagedGraphId=stagedGraphId)
    




"""The client module is used for connecting to Rendered.ai's Platform API."""

envs = {
    'prod': {
        'name': 'Rendered.ai Platform',
        'url':  'https://deckard.rendered.ai',
        'api':  'https://api.rendered.ai/graphql' },
    'test': {
        'name': 'Rendered.ai Test Platform',
        'url':  'https://deckard-test.web.app',
        'api':  'https://api.test.rendered.ai/graphql' },
    'dev': {
        'name': 'Rendered.ai Development Platform',
        'url':  'https://deckard-dev-8eaa5.web.app',
        'api':  'https://api.dev.rendered.ai/graphql' },
    'infra': {
        'name': 'Rendered.ai Infrastructure Platform',
        'url':  'https://deckard-infra.web.app',
        'api':  'https://api.infra.rendered.ai/graphql' }
}


class client:

    def __init__(self, workspaceId=None, environment='prod', email=None, password=None, local=False, interactive=True, verbose=None):
        import getpass
        import time
        import os
        from anatools.anaclient.api import api
        self.verbose = verbose
        self.interactive = interactive
        if environment not in ['dev','test','prod','infra']:  print("Invalid environment argument, must be 'infra', 'dev', 'test' or 'prod'."); return
        if local:
            os.environ['NO_PROXY'] = '127.0.0.1'
            self.__url = 'http://127.0.0.1:3000'
            if interactive: print("Local is set to",self.__url)
        else: self.__url = envs[environment]['api']
        self.__password = password
        self.__logout = True
        self.email = email
        self.user = None
        self.organizations = {}
        self.workspaces ={}
        self.channels = {}
        self.volumes = {}
        self.organization = None

        self.ana_api = api(self.__url, None, self.verbose)
        if not self.email:
            print(f'Enter your credentials for the {envs[environment]["name"]}.') 
            self.email = input('Email: ')
        if not self.__password:
            failcount = 1
            while self.user is None:
                self.__password = getpass.getpass()
                self.user = self.ana_api.login(self.email, self.__password)
        if self.user is None:
            try: self.user = self.ana_api.login(self.email, self.__password)
            except: print(f'Failed to login to {envs[environment]["name"]} with email {self.email}.'); return
        if self.user is False: print(f'Failed to login to {envs[environment]["name"]} with email {self.email}.'); return
        if self.verbose == 'debug': print(f'{self.user["uid"]}\n{self.user["idtoken"]}')
        self.__logout = False
        self.ana_api = api(self.__url, {'uid':self.user['uid'], 'idtoken':self.user['idtoken']}, self.verbose)
        self.get_organizations()
        if len(self.organizations) == 0: print("No organizations available. Contact support@rendered.ai for support or fill out a form at https://rendered.ai/#contact."); return
        self.get_workspaces()
        if len(self.workspaces) == 0: 
            self.workspace = None
            print("No workspaces available. Contact support@rendered.ai for support or fill out a form at https://rendered.ai/#contact."); 
            return
        self.get_channels()
        self.get_volumes()
        if workspaceId:     
            self.workspace = workspaceId
            for workspace in self.workspaces:
                if self.workspace == workspace['workspaceId']: self.organization = workspace['organizationId']
            if self.organization is None:
                print("The workspaceId provided is invalid. If you believe this is a mistake, contact support@rendered.ai for support or fill out a form at https://rendered.ai/#contact.")
                for workspace in self.workspaces: print(workspace["workspaceId"])
                self.workspace = None
                return
        else:
            self.workspace = self.workspaces[0]['workspaceId']
            self.organization = self.workspaces[0]['organizationId']
            if interactive: 
                print(f'These are your organizations and workspaces:')
                for organization in self.organizations:
                    print(f"    {organization['name']+' Organization'[:44]:<44}  {organization['organizationId']:<50}")
                    for workspace in self.workspaces:
                        if workspace["organizationId"] == organization["organizationId"]:
                            print(f"\t{workspace['name'][:40]:<40}  {workspace['workspaceId']:<50}")
        if interactive: 
            print(f'Signed into {envs[environment]["name"]} with {self.email}')
            print(f'The current organization is: {self.organization}')
            print(f'The current workspace is: {self.workspace}')


    def refresh_token(self):
        import time
        from anatools.anaclient.api import api
        if int(time.time()) > int(self.user['expires']):
            self.user = self.ana_api.login(self.email, self.__password)
            self.ana_api = api(self.__url, {'uid': self.user['uid'], 'idtoken': self.user['idtoken']}, self.verbose)


    def check_logout(self):
        if self.__logout: print('You are currently logged out, login to access the Rendered.ai Platform.'); return True
        self.refresh_token()
        return False


    def logout(self):
        """Logs out of the ana sdk and removes credentials from ana."""
        if self.check_logout(): return
        self.__logout = True
        del self.__password, self.__url, self.user


    def login(self, workspaceId=None, environment='prod', email=None, password=None, local=False, interactive=True, verbose=None):
        """Log in to the SDK. 
        
        Parameters
        ----------
        workspaceId: str
            ID of the workspace to log in to. Uses default if not specified.
        environment: str
            Environment to log into. Defaults to production.
        email: str
            Email for the login. Will prompt if not provided.
        password: str
            Password to login. Will prompt if not provided.
        local: bool
            Used for development to indicate pointing to local API.
        verbose: str
            Flag to turn on verbose logging. Use 'debug' to view log output.
        
        """
        self.__init__(workspaceId, environment, email, password, local, interactive, verbose)

    
    from .organizations import get_organization, set_organization, get_organizations, edit_organization, get_organization_members, get_organization_invites, add_organization_member, edit_organization_member, remove_organization_member, remove_organization_invitation ,get_organization_limits, set_organization_limit, get_organization_usage
    from .workspaces    import get_workspace, set_workspace, get_workspaces, create_workspace, edit_workspace, delete_workspace, get_workspace_guests, get_workspace_invites, add_workspace_guest, remove_workspace_guest, remove_workspace_invitation, get_workspace_limits, set_workspace_limit
    from .graphs        import get_staged_graphs, create_staged_graph, edit_staged_graph, delete_staged_graph, download_staged_graph, get_default_graph, set_default_graph
    from .datasets      import get_datasets, create_dataset, edit_dataset, delete_dataset, download_dataset, cancel_dataset, upload_dataset, get_dataset_runs, get_dataset_log
    from .channels      import get_channels, get_managed_channels, create_managed_channel, edit_managed_channel, delete_managed_channel, add_channel_access, remove_channel_access, build_managed_channel, deploy_managed_channel, get_deployment_status, get_channel_documentation, upload_channel_documentation
    from .volumes       import get_volumes, get_managed_volumes, create_managed_volume, edit_managed_volume, delete_managed_volume, add_volume_access, remove_volume_access, get_volume_data, download_volume_data, upload_volume_data, delete_volume_data, mount_volumes
    from .analytics     import get_analytics, get_analytics_types, create_analytics, delete_analytics
    from .annotations   import get_annotations, get_annotation_formats, get_annotation_maps, create_annotation, download_annotation, delete_annotation 
    from .gan           import get_gan_models, get_gan_dataset, create_gan_dataset, delete_gan_dataset, upload_gan_model, delete_gan_model, add_gan_access, remove_gan_access
    from .umap          import get_umap, create_umap, delete_umap



# Copyright 2019-2022 DADoES, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License in the root directory in the "LICENSE" file or at:
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import sys
import logging
from anatools.lib.node import Node

logger = logging.getLogger(__name__)


class ConditionalSelector(Node):

    def exec(self):
        logger.info("Executing {}".format(self.name))
        # parse the inputs
        value = None
        a = float(self.inputs['ConditionA'][0])
        operator = str(self.inputs['Operator'][0])
        b = float(self.inputs['ConditionB'][0])
        t = self.inputs['True'][0]
        f = self.inputs['False'][0]
        if operator == "Less Than":
            if a < b:   value = t
            else:       value = f
        elif operator == "Equal To":
            if a == b:  value = t
            else:       value = f
        elif operator == "Greater Than":
            if a > b:   value = t
            else:       value = f
        else:
            logger.error("Encountered invalid value for Operator: {}. Exiting...".format(operator))
            sys.exit(1)
        return {"Value": value}



"""
Channels API calls.
"""

def getChannels(self, organizationId=None, workspaceId=None, channelId=None, verbose=False):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "getChannels",
            "variables": {
                "organizationId": organizationId,
                "workspaceId": workspaceId,
                "channelId": channelId
            },
            "query": """query 
                getChannels($organizationId: String, $workspaceId: String, $channelId: String) {
                    getChannels(organizationId: $organizationId, workspaceId: $workspaceId, channelId: $channelId) {
                        channelId
                        organizationId
                        name
                        updatedAt
                    }
                }"""})
    return self.errorhandler(response, "getChannels")


def getManagedChannels(self, organizationId, channelId=None):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "getManagedChannels",
            "variables": {
                "organizationId": organizationId,
                "channelId": channelId
            },
            "query": """query 
                getManagedChannels($organizationId: String, $channelId: String) {
                    getManagedChannels(organizationId: $organizationId, channelId: $channelId) {
                        channelId
                        organizationId
                        name
                        instanceType
                        volumes
                        timeout
                        interfaceVersion
                        createdAt
                        updatedAt
                        organizations {
                            organizationId
                            name
                        }
                    }
                }"""})
    return self.errorhandler(response, "getManagedChannels")


def getChannelDeployment(self, deploymentId):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "getChannelDeployment",
            "variables": {
                "deploymentId": deploymentId
            },
            "query": """query 
                getChannelDeployment($deploymentId: String!) {
                    getChannelDeployment(deploymentId: $deploymentId) {
                        deploymentId
                        channelId
                        status {
                            state
                            step
                            message
                        }
                        createdBy
                        createdAt
                        updatedAt
                    }
                }"""})
    return self.errorhandler(response, "getChannelDeployment")


def createManagedChannel(self, organizationId, name, volumes=None, instance=None, timeout=None, interfaceVersion=None):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "createManagedChannel",
            "variables": {
                "organizationId": organizationId,
                "name": name,
                "volumes": volumes,
                "instance": instance,
                "timeout": timeout,
                "interfaceVersion": interfaceVersion
            },
            "query": """mutation 
                createManagedChannel($organizationId: String!, $name: String!, $volumes: [String], $instance: String, $timeout: Int, $interfaceVersion: Int) {
                    createManagedChannel(organizationId: $organizationId, name: $name, volumes: $volumes, instance: $instance, timeout: $timeout, interfaceVersion: $interfaceVersion)
                }"""})
    return self.errorhandler(response, "createManagedChannel")


def deleteManagedChannel(self, channelId):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "deleteManagedChannel",
            "variables": {
                "channelId": channelId
            },
            "query": """mutation 
                deleteManagedChannel($channelId: String!) {
                    deleteManagedChannel(channelId: $channelId)
                }"""})
    return self.errorhandler(response, "deleteManagedChannel")


def editManagedChannel(self, channelId, name=None, volumes=None, instance=None, timeout=None, status=None, interfaceVersion=None):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "editManagedChannel",
            "variables": {
                "channelId": channelId,
                "name": name,
                "volumes": volumes,
                "instance": instance,
                "timeout": timeout,
                "status":status,
                "interfaceVersion": interfaceVersion
            },
            "query": """mutation 
                editManagedChannel($channelId: String!, $name: String, $volumes: [String], $instance: String, $timeout: Int, $status: String, $interfaceVersion: Int) {
                    editManagedChannel(channelId: $channelId, name: $name, volumes: $volumes, instance: $instance, timeout: $timeout, status: $status, interfaceVersion: $interfaceVersion)
                }"""})
    return self.errorhandler(response, "editManagedChannel")


def addChannelOrganization(self, channelId, organizationId):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "addChannelOrganization",
            "variables": {
                "channelId": channelId,
                "organizationId": organizationId
            },
            "query": """mutation 
                addChannelOrganization($channelId: String!, $organizationId: String!) {
                    addChannelOrganization(channelId: $channelId, organizationId: $organizationId)
                }"""})
    return self.errorhandler(response, "addChannelOrganization")


def removeChannelOrganization(self, channelId, organizationId):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "removeChannelOrganization",
            "variables": {
                "channelId": channelId,
                "organizationId": organizationId
            },
            "query": """mutation 
                removeChannelOrganization($channelId: String!, $organizationId: String!) {
                    removeChannelOrganization(channelId: $channelId, organizationId: $organizationId)
                }"""})
    return self.errorhandler(response, "removeChannelOrganization")


def deployManagedChannel(self, channelId, alias=None):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "deployManagedChannel",
            "variables": {
                "channelId": channelId,
                "alias": alias
            },
            "query": """mutation 
                deployManagedChannel($channelId: String!, $alias: String) {
                    deployManagedChannel(channelId: $channelId, alias: $alias) {
                        deploymentId
                        ecrEndpoint
                        ecrPassword
                    }
                }"""})
    return self.errorhandler(response, "deployManagedChannel")


def setChannelGraph(self, channelId, workspaceId, graphId=None, stagedGraphId=None):
    response = self.session.post(
        url = self.url, 
        headers = self.headers,
        json = {
            "operationName": "setChannelGraph",
            "variables": {
                "channelId": channelId,
                "workspaceId": workspaceId,
                "graphId": graphId,
                "stagedGraphId": stagedGraphId
            },
            "query": """mutation 
                setChannelGraph($channelId: String!, $workspaceId: String!, $graphId: String, $stagedGraphId: String) {
                    setChannelGraph(channelId: $channelId, workspaceId: $workspaceId, graphId: $graphId, stagedGraphId: $stagedGraphId)
                }"""})
    return self.errorhandler(response, "setChannelGraph")

    
def getChannelDocumentation(self, channelId):
    response = self.session.post(
        url = self.url, 
        headers = self.headers,
        json = {
            "operationName": "getChannelDocumentation",
            "variables": {
                "channelId": channelId
            },
            "query": """query getChannelDocumentation($channelId: String!) {
                getChannelDocumentation(channelId: $channelId) {
                    markdown
                }
            }"""})
    return self.errorhandler(response, "getChannelDocumentation")


def uploadChannelDocumentation(self, channelId, keys=[]):
    response = self.session.post(
        url = self.url, 
        headers = self.headers,
        json = {
            "operationName": "uploadChannelDocumentation",
            "variables": {
                "channelId": channelId,
                "keys": keys,
            },
            "query": """mutation uploadChannelDocumentation($channelId: String!, $keys: [String]!) {
                uploadChannelDocumentation(channelId: $channelId, keys: $keys) {
                    key
                    url
                    fields {
                        key
                        bucket
                        algorithm
                        credential
                        date
                        token
                        policy
                        signature
                    }
                }
            }"""})
    return self.errorhandler(response, "uploadChannelDocumentation")


"""API Module"""

class api:
    
    def __init__(self, url, headers, verbose=False):
        import requests
        self.url = url
        self.headers = headers
        self.verbose = verbose
        self.session = requests.Session()

    def login(self, email, password):
        response = self.session.post(
            url = self.url, 
            json = {
                "operationName": "signIn",
                "variables": {
                    "email": email,
                    "password": password
                },
                "query": """mutation 
                    signIn($email: String!, $password: String!) {
                        signIn(email: $email, password: $password) {
                            uid
                            idtoken
                            expires
                        }
                    }"""})
        if 'errors' in response.json(): return False
        return self.errorhandler(response, "signIn")

    def close(self):
        self.session.close()

    from .handlers      import errorhandler
    from .organizations import getOrganizations, editOrganization
    from .channels      import getChannels, getManagedChannels, getChannelDeployment, createManagedChannel, deleteManagedChannel, editManagedChannel, addChannelOrganization, removeChannelOrganization, deployManagedChannel, setChannelGraph, getChannelDocumentation, uploadChannelDocumentation
    from .volumes       import getVolumes, getManagedVolumes, createManagedVolume, deleteManagedVolume, editManagedVolume, addVolumeOrganization, removeVolumeOrganization, getVolumeData, putVolumeData, deleteVolumeData, mountVolumes
    from .members       import getMembers, addMember, removeMember, editMember, getInvitations
    from .workspaces    import getWorkspaces, createWorkspace, deleteWorkspace, editWorkspace
    from .graphs        import getGraphs, createGraph, deleteGraph, editGraph, downloadGraph, getDefaultGraph
    from .datasets      import getDatasets, createDataset, deleteDataset, editDataset, downloadDataset, cancelDataset, datasetUpload, getDatasetRuns, getDatasetLog
    from .limits        import getPlatformLimits, setPlatformLimit, getOrganizationLimits, setOrganizationLimit, getWorkspaceLimits, setWorkspaceLimit, getOrganizationUsage
    from .analytics     import getAnalytics, getAnalyticsTypes, createAnalytics, deleteAnalytics
    from .annotations   import getAnnotations, getAnnotationFormats, getAnnotationMaps, createAnnotation, downloadAnnotation
    from .gan           import getGANModels, getGANDataset, createGANDataset, deleteGANDataset, uploadGANModel, deleteGANModel, addGANAccess, removeGANAccess
    from .umap          import getUMAP, createUMAP, deleteUMAP



"""
Analytics Functions
"""

def get_analytics_types(self):
    """Retrieve the analytics types available on the Platform.
    
    Returns
    -------
    list[str]
        The analytics types available on the Platform.
    """
    if self.check_logout(): return
    return self.ana_api.getAnalyticsTypes()


def get_analytics(self, datasetId, analyticsId, workspaceId=None):
    """Retrieve information about analytics jobs. Images will get downloaded to current working directory if available. 
    
    Parameters
    ----------
    datasetId : str
        Dataset ID that the analytics job is associated with.
    analyticsId : str
        Job ID for an analytics job. 
    workspaceId: str
        Workspace ID where the analytics exist. If none is provided, the current workspace will get used. 
    
    Returns
    -------
    list[dict]
        Analytics job information.
    """
    import os, json, requests
    if self.check_logout(): return
    if datasetId is None: raise ValueError("DatasetId must be provided.")
    if analyticsId is None: raise ValueError("AnalyticsId must be provided.")
    if workspaceId is None: workspaceId = self.workspace
    result = self.ana_api.getAnalytics(workspaceId=workspaceId, datasetId=datasetId, analyticsId=analyticsId)
    analytics_result = json.dumps(result)
    presigned_urls = [str for str in analytics_result.split("\\\"") if str.startswith("https")]

    for i in range(len(presigned_urls)):
        filename = presigned_urls[i].split("/")[-1].split("?")[0]
        print(f"\r[{i+1} / {len(presigned_urls)}]  Downloading {filename}...", end="", flush=True)
        
        with requests.get(presigned_urls[i], allow_redirects=True) as response:
            with open(filename, "wb") as outfile:
                outfile.write(response.content)
        analytics_result = analytics_result.replace(presigned_urls[i], os.path.join(os.getcwd(), filename), 1)
    print("\r", end="")

    return json.loads(analytics_result)


def create_analytics(self, datasetId, type, range=[], images=True, workspaceId=None):
    """Generate analytics for a dataset.
    
    Parameters
    ----------
    datasetId : str
        Dataset ID to download image annotation for.
    type : str
        The type of analytics to generate. Choose one from the list that `get_analytics_types` method returns.  
    range : list[int]
        The range of runs to generate analytics for.
    images : bool
        If true, images specific to the analytics type will be created along with metrics data. 
    workspaceId : str
        Workspace ID of the dataset to generate the analytics for. If none is provided, the current workspace will get used. 
    
    Returns
    -------
    str
        The analyticsId for the analytics job.
    """
    if self.check_logout(): return
    if workspaceId is None: workspaceId = self.workspace
    return self.ana_api.createAnalytics(workspaceId=workspaceId, datasetId=datasetId, typeInput=type, rangeInput=range, returnImages=images)


def delete_analytics(self, analyticsId, workspaceId=None):
    """Deletes a dataset's analytics.
    
    Parameters
    ----------
    analyticsId : str
        Analytics ID for the analytics to delete. 
    workspaceId: str
        Workspace ID where the analytics exist. If none is provided, the current workspace will get used. 
    
    Returns
    -------
    bool
        If true, successfully deleted the analytics.
    """
    if self.check_logout(): return
    if analyticsId is None: raise ValueError("AnalyticsId must be defined.")
    if workspaceId is None: workspaceId = self.workspace
    return self.ana_api.deleteAnalytics(workspaceId=workspaceId, analyticsId=analyticsId)




# Copyright 2019-2022 DADoES, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License in the root directory in the "LICENSE" file or at:
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import yaml
import glob
import logging
import os
import importlib
from pathlib import Path
import anatools
import anatools.lib.context as ctx

logger = logging.getLogger(__name__)


def find_channelfile():
    channel = None
    channelfiles = []
    # search local directory
    localfiles = [file for file in os.listdir('.') if file.endswith('.yml') or file.endswith('.yaml') ]
    for file in localfiles: channelfiles.append(file)
    # search workspace directory (devcontainer)
    if os.path.exists('/workspaces'):
        workspaces = os.listdir('/workspaces')
        for workspace in workspaces: 
            workspacefiles = [f'/workspaces/{workspace}/{file}' for file in os.listdir(f'/workspaces/{workspace}') if file.endswith('.yml') or file.endswith('.yaml') ]
            for file in workspacefiles: channelfiles.append(file)
    # search ana directory (deployed)
    if os.path.exists('/ana'):
        anafiles = [f'/ana/{file}' for file in os.listdir(f'/ana') if file.endswith('.yml') or file.endswith('.yaml') ]
        for file in anafiles: channelfiles.append(file)
    for channelfile in channelfiles:
        try: 
            channel = Channel(channelfile)
            if len(channel.classes.keys()) == 0: raise Exception
            channel = channelfile
            print(f'Using channelfile found at {channel}.\nIf this is the wrong channel, specify a channelfile using the --channel argument.')
            break
        except: pass
    return channel

def get_schema_files(package_name):
    ''' Return a list of all schema files in the package '''
    package = importlib.import_module(package_name)
    package_init = package.__file__
    if package_init is None:
        logger.critical(f"Package {package.__name__} is missing __init__.py")
        raise ValueError
    package_dir = os.path.dirname(os.path.realpath(package_init))
    node_dir = os.path.join(package_dir, "nodes")
    schema_files = []
    if os.path.isdir(node_dir):
        schema_files = glob.glob(os.path.join(node_dir, "*.yml"))
    return schema_files

class Channel:
    console_logging_configured = False
    logfile_logging_configured = False
    def __init__(self, channel_file):
        """ Create a channel class from a channel file """
        self.classes = {}
        self.schemas = {}
        self.packages = {}
        self.type = "blender" # assume it's a blender channel unless otherwise specified

        # get directory where ana is installed
        ana_init = anatools.__file__
        self.ana_package_dir = os.path.dirname(os.path.realpath(ana_init))

        # default execution flags
        self.execution_flags = {
            "--channel": "channel.yml",
            "--graph": "graphs/default.yml",
            "--loglevel": "ERROR",
            "--logfile": None,
            "--seed": None,
            "--interp_num": 0,
            "--preview": False,
            "--output": "./output",
            "--data": "./data"
        }

        # default channel settings
        self.channel_settings = {
            "type": "blender"
        }

        # get an ordered list of the channel files to load
        done = False
        channel_file_list = [channel_file]
        while not done:
            with open(channel_file, 'r') as f:
                cfg = yaml.safe_load(f)

            if "channel" in cfg and "base" in cfg["channel"]:
                channel_file = cfg["channel"]["base"]
                channel_file_list.append(channel_file)
            else:
                done = True
        channel_file_list.reverse()

        # process channel files in order
        channel_settings = {
            "type": "blender",
        }
        add_setup_modules = []
        remove_setup_modules = []
        add_packages = []
        remove_packages = []
        add_nodes = []
        remove_nodes = []
        rename_nodes = []
        for channel_file in channel_file_list:
            with open(channel_file, 'r') as f:
                cfg = yaml.safe_load(f)
            if "channel" in cfg:
                channel_settings = {**channel_settings, **cfg["channel"]}
            if "add_setup" in cfg:
                add_setup_modules.extend(cfg["add_setup"])
            if "remove_setup" in cfg:
                remove_setup_modules.extend(cfg["remove_setup"])
            if "add_packages" in cfg:
                add_packages.extend(cfg["add_packages"])
            if "remove_packages" in cfg:
                remove_packages.extend(cfg["remove_packages"])
            if "add_nodes" in cfg:
                add_nodes.extend(cfg["add_nodes"])
            if "remove_nodes" in cfg:
                remove_nodes.extend(cfg["remove_nodes"])
            if "rename_nodes" in cfg:
                rename_nodes.extend(cfg["rename_nodes"])
            if "default_execution_flags" in cfg:
                # channel file defaults take precedence
                self.execution_flags = {**self.execution_flags, **cfg["default_execution_flags"]}

        # save channel settings
        self.type = channel_settings["type"]

        # remove setup modules
        for remove_setup_module in remove_setup_modules:
            found = False
            for i, setup_module in enumerate(add_setup_modules):
                if remove_setup_module == setup_module:
                    add_setup_modules.pop(i)
                    found = True
                    break
            if not found:
                logger.critical(f"Can't remove setup module {remove_setup_module}. Not found.")
                raise ValueError
        self.setup_modules = add_setup_modules

        # remove packages
        for remove_package in remove_packages:
            found = False
            for i, add_package in enumerate(add_packages):
                if remove_package == add_package:
                    add_packages.pop(i)
                    found = True
                    break
            if not found:
                logger.critical(f"Can't remove package {remove_package}. Not found.")
                raise ValueError
        
        # import packages
        package_list = []
        for package in add_packages:
            if package not in package_list:
                package_list.append(package)
                schema_files = get_schema_files(package)
                
                # load schemas and save class info
                for schema_file in schema_files:
                    try:
                        with open(schema_file, "r") as f:
                            schema_module = yaml.safe_load(f)
                    except:
                        logger.critical(f"Error reading schema '{schema_file}'")
                        raise
                    python_module = package + ".nodes." + Path(schema_file).stem
                    for schema in schema_module["schemas"]:
                        alias = schema_module["schemas"][schema].get("alias", schema)
                        self.schemas[alias] = schema_module["schemas"][schema]
                        self.classes[alias] = {
                            "module": python_module,
                            "class": schema
                    }

        # add nodes
        for node_dict in add_nodes:
            node_name = node_dict["name"]
            package = node_dict["package"]
            if package not in package_list:
                package_list.append(package)
            alias = node_dict.get("alias", None)
            category = node_dict.get("category", None)
            subcategory = node_dict.get("subcategory", None)
            color = node_dict.get("color", None)
            # get all the schema files in the package
            schema_files = get_schema_files(package)
            if len(schema_files) == 0:
                logger.critical(f"Schema for '{node_name}' not found")
                raise ValueError
            
            # search all schemas in the package for the node name
            found = False
            for schema_file in schema_files:
                try:
                    with open(schema_file, "r") as f:
                        schema_module = yaml.safe_load(f)
                except:
                    logger.critical(f"Error reading schema '{schema_file}'")
                    raise
                for schema in schema_module["schemas"]:
                    if schema == node_name:
                        python_module = package + ".nodes." + Path(schema_file).stem
                        # check if schema file included an alias
                        original_alias = schema_module["schemas"][schema].get("alias", node_name)
                        if alias is None: alias = original_alias
                        if category: schema_module["schemas"][schema]['category'] = category
                        if subcategory: schema_module["schemas"][schema]['subcategory'] = subcategory
                        if color: schema_module["schemas"][schema]['color'] = color
                        self.schemas[alias] = schema_module["schemas"][schema]
                        self.classes[alias] = {
                            "module": python_module,
                            "class": schema
                        }
                        found = True
                        break
                if found:
                    break
            if not found:
                logger.critical(f"Schema for '{node_name}' not found")
                raise ValueError

        # remove nodes
        for node_name in remove_nodes:
            self.schemas.pop(node_name)
            self.classes.pop(node_name)

        # rename node
        for rename_dict in rename_nodes:
            old_name = rename_dict["old_name"]
            new_name = rename_dict["new_name"]
            if old_name not in self.schemas:
                logger.critical(f"Can't rename node - '{old_name}' not found.")
                raise ValueError
            self.schemas[new_name] = self.schemas.pop(old_name)
            self.classes[new_name] = self.classes.pop(old_name)

        # read package configurations
        for package_name in package_list:
            package = importlib.import_module(package_name)
            package_init = package.__file__
            package_dir = os.path.dirname(os.path.realpath(package_init))
            package_config_file = os.path.join(package_dir, "package.yml")
            if os.path.exists(package_config_file):
                with open(package_config_file, "r") as f:
                    self.packages[package_name] = yaml.safe_load(f)
            else:
                self.packages[package_name] = {}

        # construct the node menu
        if "node_menu" in cfg:
            self.node_menu = cfg["node_menu"]
        else:
            self.node_menu = None

        self.name = os.path.splitext(os.path.basename(channel_file))[0]

    @classmethod
    def configure_logging(cls, loglevel="ERROR", logfile=None, logfile_mode="w"):

        # set up logging
        rootLogger = logging.getLogger()
        numeric_level = getattr(logging, loglevel.upper(), None)
        if not isinstance(numeric_level, int):
            raise ValueError('Invalid log level: {}'.format(loglevel))
        rootLogger.setLevel(numeric_level)
        formatter = logging.Formatter(
            '%(asctime)s %(name)s %(levelname)s - %(message)s',
            datefmt='%Y-%m-%d %H:%M:%S')

        # log to file
        if not Channel.logfile_logging_configured and logfile is not None:
            fileHandler = logging.FileHandler(logfile, mode=logfile_mode)
            fileHandler.setFormatter(formatter)
            rootLogger.addHandler(fileHandler)
            Channel.logfile_logging_configured = True

        # log to console
        if not Channel.console_logging_configured:
            consoleHandler = logging.StreamHandler()
            consoleHandler.setFormatter(formatter)
            rootLogger.addHandler(consoleHandler)
            Channel.console_logging_configured = True


    def process_args(self, args):
        if args.channel is not None:
            self.execution_flags["--channel"] = args.channel
        if args.graph is not None:
            self.execution_flags["--graph"] = args.graph
        if args.loglevel is not None:
            self.execution_flags["--loglevel"] = args.loglevel
        if args.logfile is not None:
            self.execution_flags["--logfile"] = args.logfile
        if args.seed is not None:
            self.execution_flags["--seed"] = args.seed
        if args.interp_num is not None:
            self.execution_flags["--interp_num"] = args.interp_num
        if args.preview is not None:
            self.execution_flags["--preview"] = args.preview
        if args.output is not None:
            self.execution_flags["--output"] = args.output
        if args.data is not None:
            self.execution_flags["--data"] = args.data
        
        # Configure logging
        Channel.configure_logging(
            loglevel=self.execution_flags["--loglevel"],
            logfile=self.execution_flags["--logfile"],
            logfile_mode="a"
        )

        # Configure global variables
        ctx.initialize(
            channel=self,
            seed=self.execution_flags["--seed"],
            interp_num=self.execution_flags["--interp_num"],
            preview=self.execution_flags["--preview"],
            output=self.execution_flags["--output"],
            data=self.execution_flags["--data"])

    def setup(self):
        """ Execute channel setup modules """
        for setup_module in self.setup_modules:
            setup_method = getattr(importlib.import_module(setup_module), "setup")
            setup_method()


"""
Channels Functions
"""

def get_channels(self, organizationId=None, workspaceId=None, channelId=None):
    """Shows all channels available to the user. Can filter by organizationId, workspaceId, or channelId.
    
    Parameters
    ----------
    organizationId : str
        Filter channel list on what's available to the organization.
    workspaceId : str    
        Filter channel list on what's available to the workspace.
    channelId: str
        Filter channel list on the specific channelId.
    
    Returns
    -------
    list[dict]
        List of channels associated with user, workspace, organization or channelId.
    """
    if self.check_logout(): return
    channels = self.ana_api.getChannels(organizationId=organizationId, workspaceId=workspaceId, channelId=channelId)
    if channels:
        for channel in channels:
            self.channels[channel['channelId']] = channel['name']
    return channels
       
                

def get_managed_channels(self, channelId=None, organizationId=None):
    """Get information for all managed channels that you own within your organization.
    
    Parameters
    ----------
    channelId : str
        Channel Id to filter.
    organizationId : str
        Organization ID. Defaults to current if not specified.
   
    Returns
    -------
    list[dict]
        channel data
    """
    if self.check_logout(): return
    if organizationId is None: organizationId = self.organization
    return self.ana_api.getManagedChannels(organizationId=organizationId, channelId=channelId)


def create_managed_channel(self, name, organizationId=None, volumes=[], instance='p2.xlarge', timeout=120, interfaceVersion=1):
    """Create a managed channel for your organization.
    
    Parameters
    ----------
    name : str
        Channel name.
    organizationId : str
        Organization ID. Defaults to current if not specified.
    volumes : list[str]
        List of the data volume names to associate with this channel.
    instance: str
        AWS Instance type.
    timeout: int
        Maximum runtime of a channel run.
    interface: int
        The ana interface version number.
   
    Returns
    -------
    list[dict]
        channel data
    """
    if self.check_logout(): return
    if organizationId is None: organizationId = self.organization
    result = self.ana_api.createManagedChannel(organizationId=organizationId, name=name, volumes=volumes, instance=instance, timeout=timeout, interfaceVersion=interfaceVersion)
    self.get_channels()
    return result

def edit_managed_channel(self, channelId, name=None, volumes=None, instance=None, timeout=None, status=None, interfaceVersion=None):
    """Edit a managed channel for your organization.
    
    Parameters
    ----------
    channelId : str
        ChannelId ID of the channel to edit.
    name : name
        The new name to give the channel.
    volumes : list[str]
        Data volumes for the channel.
    instance: str
        Instance type to run the channel on.
    timeout: int
        Maximum runtime for the channel run.
    status: str
        The status of the channel.
    interface: int
        The ana interface version number.
    
    Returns
    -------
    bool
        If true, the channel was successfully edited.
    """
    if self.check_logout(): return
    if channelId is None: raise Exception('ChannelId must be specified.')
    result = self.ana_api.editManagedChannel(channelId=channelId, name=name, volumes=volumes, instance=instance, timeout=timeout, status=status, interfaceVersion=interfaceVersion)
    self.get_channels()
    return result

def delete_managed_channel(self, channelId):
    """Delete a managed channel of your organization.
    
    Parameters
    ----------
    channelId : str
        Id of channel to delete.
    
    Returns
    -------
    str
        Status
    """
    if self.check_logout(): return
    if channelId is None: raise Exception('ChannelId must be specified.')
    result = self.ana_api.deleteManagedChannel(channelId=channelId)
    self.get_channels()
    return result
    
def add_channel_access(self, channelId, organizationId):
    """Add access to a channel for an organization.
    
    Parameters
    ----------
    channelId : str
        Id of channel to add access for.
    organizationId : str
        Organization ID. Defaults to current if not specified.
    
    Returns
    -------
    str
        Access status. 
    """
    if self.check_logout(): return
    if organizationId is None: raise Exception('OrganizationId must be specified.')
    return self.ana_api.addChannelOrganization(channelId=channelId, organizationId=organizationId)


def remove_channel_access(self, channelId, organizationId):
    """Remove access to a channel for an organization.
    
    Parameters
    ----------
    channelId : str
        Id of channel to remove access to.
    organizationId : str
        Organization ID. Defaults to current if not specified.
    
    Returns
    -------
    str
        Access status. 
    """
    if self.check_logout(): return
    if organizationId is None: raise Exception('OrganizationId must be specified.')
    return self.ana_api.removeChannelOrganization(channelId=channelId, organizationId=organizationId)


def build_managed_channel(self, channelfile):
    """Build the Docker image of a channel.
    
    Parameters
    ----------
    channelfile : str
        The channel file for the channel to build.
    Returns
    -------
    bool
        A boolean that indicates if the channel Docker image was successfully built.
    """
    import os
    import shutil
    import docker
    import time

    # build dockerfile
    print('Building Dockerfile...', end='')
    if not os.path.isfile(channelfile): raise Exception(f'No channel file {channelfile} found.')
    channeldir, channelfile = os.path.split(channelfile)
    if channeldir == "": channeldir = "./"   
    if not os.path.isdir(os.path.join(channeldir, '.devcontainer')): raise Exception(f'No .devcontainer directory found in channel directory.')
    if not os.path.isfile(os.path.join(channeldir, '.devcontainer/Dockerfile')): raise Exception(f'Issue detected with .devcontainer directory, build requires Dockerfile.')
    builddir = os.path.join(channeldir,'.devcontainer/.build')
    if os.path.isdir(builddir): shutil.rmtree(builddir)
    os.mkdir(builddir)
    with open(os.path.join(builddir, 'Dockerfile'), 'w+') as buildfile:
        with open(os.path.join(channeldir,'.devcontainer/Dockerfile')) as dockerfile: buildfile.write(dockerfile.read())
        deploycommands = """
# Commands used for deployment
RUN sudo mkdir /ana /data
COPY . /ana/
WORKDIR /ana
RUN sudo /home/$USERNAME/miniconda3/bin/pip install -r /ana/requirements.txt > /dev/null 2>&1 && \\
    sudo /home/$USERNAME/miniconda3/envs/anatools/bin/pip install -r /ana/requirements.txt > /dev/null 2>&1
USER root"""
        buildfile.write(deploycommands)
    print('done.')

    # build the tar file, include what's necesary for the container
    print('Building Docker Context...', end='')
    dirs_to_not_include = ['.devcontainer', 'data', 'output']
    files = [path for path in os.listdir(channeldir) if path not in dirs_to_not_include]
    for filepath in files:
        full_path = os.path.join(channeldir, filepath)
        is_symbolic_link = os.path.islink(full_path)
        if not is_symbolic_link:
            if os.path.isdir(full_path): shutil.copytree(full_path, os.path.join(builddir, filepath))
            else: shutil.copyfile(full_path, os.path.join(builddir, filepath))
    print('done.')

    # call the docker build command
    print('Building Docker Image...', end='', flush=True)
    dockerclient = docker.from_env()
    image, logs = dockerclient.images.build(path=builddir, tag=channelfile.split('.')[0])
    print('done.')

    if not __check_docker_image_size(image):
        return False

    # cleanup
    shutil.rmtree(builddir)
    return True

def __check_docker_image_size(image):
    """ Helper function that computes size of docker image. 
    
    Parameters
    ----------
    image : object
        Docker Image Object
    
    Returns
    -------
    bool
        True if size is less than 7.5GB. False if it's bigger.
    """
    import sys

    # limit size to 7.5GB
    size_of_image = image.attrs['VirtualSize']
    size_limit = 8053063680

    if size_of_image > size_limit:
        print('\033[91m' + 'ERROR: Docker Image size is greater than 7.5GB and cannot be uploaded.' + '\033[0m')
        print('Exiting...')
        return False

    return True  

def deploy_managed_channel(self, channelId=None, channelfile=None, image=None):
    """Deploy the Docker image of a channel.
    
    Parameters
    ----------
    channelId : str
        Channel ID that you are pushing the image to. If the channelId isn't specified, it will use the image name to lookup the channelId.
    image: str
        The Docker image name. This should match the channel name when running ana. If image is not specified, it will use the channel name for the channelId.

    Returns
    -------
    str
        deploymentId for current round of deployment or an error message if something went wrong
    """
    import os
    import docker
    import base64
    if self.check_logout(): return
    if channelId is None and image is None: print('The channelId or local image must be specified.'); return
    self.get_channels()

    if channelfile:
        if self.build_managed_channel(channelfile):
            image = os.path.split(channelfile)[1].split('.')[0]
        else:
            return False
    
    # check if channel image is in Docker
    dockerclient = docker.from_env()
    if image and channelId:
        channel = image
        if channelId not in self.channels: 
            print(f'User does not have permissions to deploy to a channel with ID \"{channelId}\" on the Rendered.ai Platform.'); return
    elif image:   
        channel = image
        channels = self.get_managed_channels()
        filteredchannels = [channel for channel in channels if channel['name'] == image]
        if len(filteredchannels) == 1: channelId = filteredchannels[0]['channelId']
        elif len(filteredchannels) == 0: print(f'User does not have permissions to deploy to a channel named \"{image}\" on the Rendered.ai Platform.'); return
        else: print('User has access to multiple channels with name \"{image}\" on the Rendered.ai Platform, please specify channelId.'); return
    else:
        if channelId in self.channels: channel = self.channels[channelId]
        else: print(f'User does not have permissions to deploy to a channel with ID \"{channelId}\" on the Rendered.ai Platform.'); return
    try: 
        channelimage = dockerclient.images.get(channel)
        if not __check_docker_image_size(channelimage):
            return False
    except docker.errors.ImageNotFound: print(f'Could not find Docker image with name \"{channel}\".'); return
    except: raise Exception('Error connecting to Docker.')

    # get repository info
    print(f'Docker image \"{channel}\" will be deployed to the \"{self.channels[channelId]}\" channel.')
    print(f"Pushing Docker Image. This could take awhile...", end='', flush=True)
    dockerinfo = self.ana_api.deployManagedChannel(channelId, image)
    deploymentId = dockerinfo['deploymentId']
    reponame = dockerinfo['ecrEndpoint']
    encodedpass = dockerinfo['ecrPassword']
    if encodedpass:
        encodedbytes = encodedpass.encode('ascii')
        decodedbytes = base64.b64decode(encodedbytes)
        decodedpass = decodedbytes.decode('ascii').split(':')[-1]
    else: print('Failed to retrieve Docker credentials from Rendered.ai platform.'); return

    # tag and push image
    channelimage.tag(reponame)
    if self.verbose:
        for line in dockerclient.images.push(reponame, auth_config={'username':'AWS', 'password':decodedpass}, stream=True, decode=True):
            if 'status' and 'progress' in line:
                    print(f"\x1b[1K\r {line['status']} \"{channel}\" image to \"{self.channels[channelId]} channel\" {line['progress']}",end='')
    else:
        largest = 0
        for line in dockerclient.images.push(reponame, auth_config={'username':'AWS', 'password':decodedpass}, stream=True, decode=True):
            if 'status' and 'progressDetail' in line:
                if 'current' and 'total' in line['progressDetail']:
                    progressDetail = line['progressDetail']
                    if progressDetail['total'] >= largest:
                        largest = progressDetail['total']
                        print(f"\x1b[1K\r[{round((progressDetail['current']/progressDetail['total']) * 100)}%] {line['status']} \"{channel}\" to \"{self.channels[channelId]}\" ",end='')
    print("\nComplete!", flush=True)
    
    # cleanup docker and update channels
    dockerclient.images.remove(reponame)
    dockerclient.close()
    if self.check_logout(): return
    self.get_channels()
    return deploymentId


def get_deployment_status(self, deploymentId, stream=False):
    """Retrieves status for a channel's deployment.
    
    Parameters
    ----------
    deploymentId: str
        The deploymentId to retrieve status for
    stream: bool
        Flag to print information to the terminal so the user can avoid constant polling to retrieve status.

    Returns
    -------
    list[dict]
        Deployment status. 
    """
    import time
    if self.check_logout(): return
    if deploymentId is None: raise Exception('DeploymentId must be specified.')
    if stream:
        UP = "\033[3F"
        CLR = "\x1B[0K"
        
        data = self.ana_api.getChannelDeployment(deploymentId=deploymentId)
        print(f"{UP}Step: {data['status']['step']}{CLR}\nState: {data['status']['state']}{CLR}\nMessage: {data['status']['message']}{CLR}\n", end='', flush=True)

        while (data['status']['state'] not in ['Channel Deployment Complete','Channel Deployment Failed']):
            time.sleep(10)
            print(f"{UP}Step: {data['status']['step']}{CLR}\nState: {data['status']['state']}{CLR}\nMessage: {data['status']['message']}{CLR}\n", end='', flush=True)
            if self.check_logout(): return
            data = self.ana_api.getChannelDeployment(deploymentId=deploymentId)
        print(f"{UP}Step: {data['status']['step']}\nState: {data['status']['state']}\nMessage: {data['status']['message']}", flush=True)
    else: return self.ana_api.getChannelDeployment(deploymentId=deploymentId)

    
def get_channel_documentation(self, channelId, localDir=None):
    """Downloads a markdown file for channel documentation.
    
    Parameters
    ----------
    channelID: str
        The channelId of the channel
    localDir: str
        The location to download the file to.

    Returns
    -------
    list[str]
        The list of filenames downloaded.
    """
    import os
    import requests

    if channelId not in self.channels: raise ValueError(f'Could not find channel with ID "{channelId}"')
    if localDir is None: localDir = './'
    elif not os.path.isdir(localDir): raise ValueError(f'Could not find directory {localDir}')

    docfile = self.ana_api.getChannelDocumentation(channelId=channelId)[0]
    downloadresponse = requests.get(url=docfile['markdown'])
    filename = os.path.join(localDir, channelId+"."+docfile['markdown'].split('?')[0].split('/')[-1].split('.')[1])
    with open(filename, 'wb') as outfile:
        outfile.write(downloadresponse.content)
    return filename


def upload_channel_documentation(self, channelId, mdfile):
    """Uploads a markdown file for channel documentation.
    
    Parameters
    ----------
    channelID: str
        The channelId of the channel
    mdfile: str
        The filepath of the markdown file used for channel documentation.

    Returns
    -------
    bool
        Success/Failure of channel documenation upload.
    """
    import os
    import requests

    if channelId not in self.channels: raise ValueError(f'Could not find channel with ID "{channelId}"')
    if not os.path.isfile(mdfile): raise ValueError(f'Could not find file {mdfile}')
    if os.path.splitext(mdfile)[1] != '.md': raise ValueError('The channel documentation file must be in markdown format with .md extension.') 
    fileinfo = self.ana_api.uploadChannelDocumentation(channelId=channelId, keys=[os.path.basename(mdfile)])[0]
    with open(mdfile, 'rb') as filebytes:
        files = {'file': filebytes}
        data = {
            "key":                  fileinfo['fields']['key'],
            "bucket":               fileinfo['fields']['bucket'],
            "X-Amz-Algorithm":      fileinfo['fields']['algorithm'],
            "X-Amz-Credential":     fileinfo['fields']['credential'],
            "X-Amz-Date":           fileinfo['fields']['date'],
            "X-Amz-Security-Token": fileinfo['fields']['token'],
            "Policy":               fileinfo['fields']['policy'],
            "X-Amz-Signature":      fileinfo['fields']['signature'],
        }
        response = requests.post(fileinfo['url'], data=data, files=files)
        if response.status_code != 204: 
            print(response.status_code)
            raise Exception('Failed to upload channel documentation file.')
    return True



"""
Organization Functions
"""

def get_organization(self):
    """Get organization id of current organization. 
    
    Returns
    -------
    str
        Organization ID of current workspace.
    """
    if self.check_logout(): return
    return self.organization


def set_organization(self, organizationId, workspaceId=None):
    """Set the organization (and optionally a workspace) to the one you wish to work in.
    
    Parameters
    ----------
    organizationId : str
        Organization ID for the organization you wish to work in.
    workspaceId : str
        Workspace ID for the workspace you wish to work in. Uses default workspace if this is not set.
    """
    if self.check_logout(): return
    if organizationId is None: raise Exception('OrganizationId must be specified.')
    workspaceSet = False
    self.workspaces = self.ana_api.getWorkspaces()
    if len(self.workspaces) == 0: 
        self.workspace = None
        print("No workspaces available. Contact support@rendered.ai for support or fill out a form at https://rendered.ai/#contact."); 
        return
        
    for workspace in self.workspaces:
        if organizationId == workspace['organizationId']:
            if workspaceId is None or workspaceId == workspace['workspaceId']:
                self.workspace = workspace['workspaceId']
                self.organization = workspace['organizationId']
                workspaceSet = True
                break
    if not workspaceSet: raise Exception('Could not find organization or workspace specified.')
    print(f'Organization set to {self.organization}.')
    print(f'Workspace set to {self.workspace}.')
    return


def get_organizations(self, organizationId=None):
    """Shows the organizations the user belongs to and the user's role in that organization.
    
    Returns
    -------
    list[dict]
        Information about the organizations you belong to. 
    """  
    if self.check_logout(): return
    if organizationId is None:
        self.organizations = self.ana_api.getOrganizations(organizationId)
        return self.organizations
    else:
        organizations = self.ana_api.getOrganizations(organizationId)
        return organizations


def edit_organization(self, name, organizationId=None):
    """Update the organization name. Uses current organization if no organizationId provided.
    
    Parameters
    ----------
    name : str
        Name to update organization to.
    organizationId : str
        Organization Id to update.
    
    Returns
    -------
    bool
        True if organization was edited successfully, False otherwise.
    """  
    if self.check_logout(): return
    if name is None: return
    if organizationId is None: organizationId = self.organization
    return self.ana_api.editOrganization(organizationId, name)


def get_organization_members(self, organizationId=None):
    """Get users of an organization optionally filtered on workspace.
    
    Parameters
    ----------
    organizationId : str
        Organization ID. Defaults to current if not specified.
    
    Returns
    -------
    list[dict]
        Information about users of an organization.
    """
    if self.check_logout(): return
    if organizationId is None: organizationId = self.organization
    return self.ana_api.getMembers(organizationId=organizationId, workspaceId=None)

def get_organization_invites(self, organizationId=None):
    """Get invitations of an organization optionally filtered on workspace.
    
    Parameters
    ----------
    organizationId : str
        Organization ID. Defaults to current if not specified.
    
    Returns
    -------
    list[dict]
        Information about invitations of an organization.
    """
    if self.check_logout(): return
    if organizationId is None: organizationId = self.organization
    return self.ana_api.getInvitations(organizationId=organizationId, workspaceId=None)


def add_organization_member(self, email, role, organizationId=None):
    """Add a user to an existing organization.
    
    Parameters
    ----------
    email: str
        Email of user to add.
    role : str
        Role for user. 
    organizationId : str
        Organization ID to add members too. Uses current if not specified.
    
    Returns
    -------
    str
        Response status if user got added to workspace succesfully. 
    """
    if self.check_logout(): return
    if email is None: raise ValueError("Email must be provided.")
    if role is None: raise ValueError("Role must be provided.")
    if organizationId is None: organizationId = self.organization
    return self.ana_api.addMember(email=email, role=role, organizationId=organizationId, workspaceId=None)


def remove_organization_member(self, email, organizationId=None):
    """Remove a member from an existing organization.
    
    Parameters
    ----------
    email : str
        Member email to remove.
    organizationId: str
        Organization ID to remove member from. Removes from current organization if not specified.
    
    Returns
    -------
    str
        Response status if member got removed from organization succesfully. 
    """
    if self.check_logout(): return
    if email is None: raise ValueError("Email must be provided.")
    if organizationId is None: organizationId = self.organization
    return self.ana_api.removeMember(email=email, organizationId=organizationId, workspaceId=None)

def remove_organization_invitation(self, email, organizationId=None, invitationId=None ):
    """Remove a invitation from an existing organization.
    
    Parameters
    ----------
    email : str
        Invitation email to remove.
    organizationId: str
        Organization ID to remove member from. Removes from current organization if not specified.
    invitationId: str
        Invitation ID to remove invitation from. Removes from current organization if not specified.
    
    Returns
    -------
    str
        Response status if member got removed from organization succesfully. 
    """
    if self.check_logout(): return
    if email is None: raise ValueError("Email must be provided.")
    if invitationId is None: raise ValueError("No invitation found.")
    if organizationId is None: organizationId = self.organization
    return self.ana_api.removeMember(email=email, organizationId=organizationId, workspaceId=None, invitationId=invitationId)


def edit_organization_member(self, email, role, organizationId=None):
    """Edit a member's role. 
    
    Parameters
    ----------
    email : str
        Member email to edit.
    role: str
        Role to assign. 
    organizationId: str
        Organization ID to remove member from. Edits member in current organization if not specified.
    
    Returns
    -------
    str
        Response if member got edited succesfully. 
    """
    if self.check_logout(): return
    if email is None: raise ValueError("Email must be provided.")
    if role is None: raise ValueError("Role must be provided.")
    if organizationId is None: organizationId = self.organization
    return self.ana_api.editMember(email=email, role=role, organizationId=organizationId)


def get_organization_limits(self, organizationId=None, setting=None):
    """Get information about Organization limits and setting.
    
    Parameters
    ----------
    organizationId : str
        Organization ID. Defaults to current if not specified.
    setting : str
        Setting name.
    
    Returns
    -------
    list[dict]
        Organization limit information.
    """
    if self.check_logout(): return
    if organizationId is None: organizationId = self.organization
    return self.ana_api.getOrganizationLimits(organizationId=organizationId, setting=setting)


def set_organization_limit(self, setting, limit, organizationId=None):
    """Set Organization limit for a setting.
    
    Parameters
    ----------
    setting : str
        Setting name.
    limit: int
        Limit for the setting.
    organizationId : str
        Organization ID. Defaults to current if not specified.
    
    Returns
    -------
    list[dict]
        Organization limit information.
    """
    if self.check_logout(): return
    if organizationId is None: organizationId = self.organization
    return self.ana_api.setOrganizationLimit(organizationId=organizationId, setting=setting, limit=limit)


def get_organization_usage(self, organizationId=None, year=None, month=None, workspaceId=None, member=None):
    """Get organization usage optionally filtered on workspace or a user. 
    
    Parameters
    ----------
    organizationId : str
        Organization ID. Defaults to current if not specified.
    year: str
        Usage year to filter on.
    month: str
        Usage month to filter on.
    workspaceId : str
        Workspace Id. Optional.
    member: str
        User email. Optional.
    
    Returns
    -------
    list[dict]
        Organization usage by channels, instanceType, and time as integer.
    """
    from datetime import datetime
    if self.check_logout(): return
    if year is None: year = str(datetime.now().year)
    if month is None: month = str(datetime.now().month)
    if organizationId is None: organizationId = self.organization
    return self.ana_api.getOrganizationUsage(organizationId=organizationId, year=year, month=month, workspaceId=workspaceId, member=member)




# Copyright 2019-2022 DADoES, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License in the root directory in the "LICENSE" file or at:
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import bpy

def create_plane(width, height):
    """
    Create a plane of size width x height located at the origin.

    Note:
        This is preferrable to bpy.ops.mesh.primitive_plane_add
        because it doesn't automatically add the plane to
        a collection.
    """

    # define dimensions
    verts = [
        (-width/2.0, -height/2.0, 0),
        (width/2.0, -height/2.0, 0),
        (width/2.0, height/2.0, 0),
        (-width/2.0, height/2.0, 0)
    ]
    faces = [(0, 1, 2, 3)]
    
    # create mesh
    mesh = bpy.data.meshes.new("plane")
    mesh.from_pydata(verts, [], faces)
    mesh.update(calc_edges=True)

    # add UV layer
    mesh.uv_layers.new()

    # create the plane
    plane = bpy.data.objects.new("plane", mesh)
    plane.location = (0, 0, 0)

    return plane



import os
import yaml
import json
from PIL import Image, ImageDraw

def convert_sagemaker_od(datadir, outdir, mapping):
    """ Generate annotations for AWS Sagemaker. Annotation jpegs will be placed in <datadir>/<outputdir>.
    
    Parameters
    ----------
    datadir : str
        Location of Rendered.ai dataset output.
    outputdir : str
        Name of directory where the results should be written.
    mapfile: str
        The map file used for annotations (YAML only).
    
    Returns
    -------

    """

    annsdir = os.path.join(datadir, "annotations")
    metadir = os.path.join(datadir, "metadata")

    # Get the image shape
    sample_image_filename = [datadir + '/images/' + imgfilename for imgfilename in os.listdir(datadir + '/images')][0]
    sample_image = Image.open(sample_image_filename)
    imgshape = [sample_image.size[0], sample_image.size[1], 3]

    sodcats = list()
    for annsfilename in os.listdir(annsdir):
        with open(os.path.join(annsdir, annsfilename), 'r') as af:
            anns = json.load(af)
        with open(os.path.join(metadir, annsfilename.replace('ana', 'metadata')), 'r') as mf:
            metadata = json.load(mf)

        soddata = dict()
        soddata['file'] = anns['filename'].split('.')[0] + '.jpeg'  # Sagemaker requires images be in jpeg format
        soddata['image_size'] = [{'width':imgshape[0],'height':imgshape[1],'depth':imgshape[2]}]
        soddata['annotations'] = list()
        for obj in metadata['objects']:
            for i, prop in enumerate(mapping['properties']):
                if eval(prop):
                    for ann in anns['annotations']:
                        if ann['id'] == obj['id']:
                            objann = ann
                            break
                    else:  # All the objects from the scene are recorded in metadata; only those in the image are annotated
                        continue

                    rai_cat_id = mapping['properties'][prop]
                    cat = mapping['classes'][rai_cat_id]
                    cat_name = cat[-1]
                    if cat_name not in sodcats: sodcats.append(cat_name)              
                    soddata['annotations'].append({
                        'class_id': sodcats.index(cat_name),
                        'left':     objann['bbox'][0],
                        'top':      objann['bbox'][1],
                        'width':    objann['bbox'][2],
                        'height':   objann['bbox'][3]
                    })
                    break

        soddata['categories'] = list()
        for cId, cName in enumerate(sodcats):
            soddata['categories'].append({'class_id':cId, 'name':cName})
    
        outfile = os.path.join(outdir, '{}.json'.format(anns['filename'].split('.')[0]))
        with open(outfile, 'w') as f:
            json.dump(soddata, f)


def convert_sagemaker_ss(datadir, outdir, mapping):
    """ Generate masks for AWS Sagemaker Semantic Segmentation. Mask pngs will be placed in <datadir>/<outputdir>.
    
    Parameters
    ----------
    datadir : str
        Location of Rendered.ai dataset output.
    outputdir : str
        Name of directory where the results should be written.
    mapfile: str
        The map file used for annotations (YAML only).
    
    Returns
    -------
    """

    annsdir = os.path.join(datadir, "annotations")
    metadir = os.path.join(datadir, "metadata")

    # Get the image shape
    sample_image_filename = [datadir + '/images/' + imgfilename for imgfilename in os.listdir(datadir + '/images')][0]
    sample_image = Image.open(sample_image_filename)
    imgshape = [sample_image.size[0], sample_image.size[1], 3]

    for annsfilename in os.listdir(annsdir):
        with open(os.path.join(annsdir, annsfilename), 'r') as af:
            anns = json.load(af)
        with open(os.path.join(metadir, annsfilename.replace('ana', 'metadata')), 'r') as mf:
            metadata = json.load(mf)
        maskimg = Image.new("L", (imgshape[0], imgshape[1]))
        draw = ImageDraw.Draw(maskimg)

        for obj in metadata['objects']:
            for prop in mapping['properties']:
                if eval(prop):
                    for ann in anns['annotations']:
                        if ann['id'] == obj['id']:
                            objann = ann
                            break
                    rai_cat_id = mapping['properties'][prop]
                    draw.polygon(objann['segmentation'][0], fill=rai_cat_id, outline=rai_cat_id)
                    break

        maskimg.save(os.path.join(outdir, f'{anns["filename"].split(".")[0]}.png'))


"""
UMAP Functions
"""
import os
import requests

def get_umap(self, umapId, workspaceId=None):
    """Retrieves information about UMAP dataset comparison from the platform.
    
    Parameters
    ----------
    umapId : str
        UMAP Job ID. 
    workspaceId : str
        Workspace ID where the datasets exists.
    
    Returns
    -------
    dict
        UMAP information.
    """
    if self.check_logout(): return
    if workspaceId is None: workspaceId = self.workspace
    return self.ana_api.getUMAP(umapId=umapId, workspaceId=workspaceId)
    

def create_umap(self, datasetIds, samples, workspaceId=None):
    """Creates a UMAP dataset comparison job on the platform.
    
    Parameters
    ----------
    datasetIds : [str]
        Dataset ID to retrieve information for. 
    samples : [int]
        Samples to take from each dataset.
    workspaceId : str
        Workspace ID where the datasets exists.
    
    Returns
    -------
    str
        The UMAP Job ID.
    """
    if self.check_logout(): return
    if len(datasetIds) != len(samples): raise ValueError("The length of datasetIds must match the length of samples.")
    for sample in samples:
        if sample < 5: raise ValueError("The number of samples must be between the range of 5 and dataset runs.")
    if workspaceId is None: workspaceId = self.workspace
    return self.ana_api.createUMAP(datasetIds=datasetIds, samples=samples, workspaceId=workspaceId)


def delete_umap(self, umapId, workspaceId=None):
    """Deletes/cancels a UMAP dataset comparison on the platform.
    
    Parameters
    ----------
    umapId : str
        UMAP Job ID. 
    workspaceId : str
        Workspace ID where the datasets exists.
    
    Returns
    -------
    bool
        Status.
    """
    if self.check_logout(): return
    if workspaceId is None: workspaceId = self.workspace
    return self.ana_api.deleteUMAP(umapId=umapId, workspaceId=workspaceId)



"""
Datasets API calls.
"""

def getDatasets(self, workspaceId, datasetId=None, name=None, email=None):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "getDatasets",
            "variables": {
                "workspaceId": workspaceId,
                "datasetId": datasetId,
                "name": name,
                "email": email
            },
            "query": """query 
                getDatasets($workspaceId: String!, $datasetId: String, $name: String, $email: String) {
                    getDatasets(workspaceId: $workspaceId, datasetId: $datasetId, name: $name, member: $email) {
                        datasetId: datasetid
                        name
                        channel
                        channelId
                        graphId: source
                        interpretations: scenarios
                        user
                        type
                        status
                        priority
                        seed
                        count
                        files
                        size
                        description
                    }
                }"""})
    return self.errorhandler(response, "getDatasets")


def createDataset(self, workspaceId, graphId, name, runs, seed, priority, description=None):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "createDataset",
            "variables": {
                "workspaceId": workspaceId,
                "graphId": graphId,
                "name": name,
                "description": description,
                "interpretations": runs,
                "seed": seed,
                "priority": priority
            },
            "query": """mutation 
                createDataset($workspaceId: String!, $graphId: String!, $name: String!, $description: String, $interpretations: Int!, $seed: Int!, $priority: Int!) {
                    createDataset(workspaceId: $workspaceId, graphId: $graphId, name: $name, description: $description, interpretations: $interpretations, seed: $seed, priority: $priority)
                }"""})
    return self.errorhandler(response, "createDataset")


def deleteDataset(self, workspaceId, datasetId):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "deleteDataset",
            "variables": {
                "workspaceId": workspaceId,
                "datasetId": datasetId
            },
            "query": """mutation 
                deleteDataset($workspaceId: String!, $datasetId: String!) {
                    deleteDataset(workspaceId: $workspaceId, datasetId: $datasetId)
                }"""})
    return self.errorhandler(response, "deleteDataset")


def editDataset(self, workspaceId, datasetId, name=None, description=None):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "editDataset",
            "variables": {
                "workspaceId": workspaceId,
                "datasetId": datasetId,
                "name": name,
                "description": description
            },
            "query": """mutation 
                editDataset($workspaceId: String!, $datasetId: String!, $name: String, $description: String) {
                    editDataset(workspaceId: $workspaceId, datasetId: $datasetId, name: $name, description: $description)
                }"""})
    return self.errorhandler(response, "editDataset")


def downloadDataset(self, workspaceId, datasetId):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "downloadDataset",
            "variables": {
                "workspaceId": workspaceId,
                "datasetId": datasetId
            },
            "query": """mutation 
                downloadDataset($workspaceId: String!, $datasetId: String!) {
                    downloadDataset(workspaceId: $workspaceId, datasetId: $datasetId)
                }"""})
    return self.errorhandler(response, "downloadDataset")


def cancelDataset(self, workspaceId, datasetId):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "cancelDataset",
            "variables": {
                "workspaceId": workspaceId,
                "datasetId": datasetId
            },
            "query": """mutation 
                cancelDataset($workspaceId: String!, $datasetId: String!) {
                    cancelDataset(workspaceId: $workspaceId, datasetId: $datasetId)
                }"""})
    return self.errorhandler(response, "cancelDataset")


def datasetUpload(self, workspaceId, description, filename):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "uploadDataset",
            "variables": {
                "workspaceId": workspaceId,
                "description": description,
                "name": filename
            },
            "query": """mutation 
                uploadDataset($workspaceId: String!, $description: String!, $name: String!) {
                    uploadDataset(workspaceId: $workspaceId, description: $description, name: $name){
                        key
                        url
                        fields {
                            key
                            bucket
                            algorithm
                            credential
                            date
                            token
                            policy
                            signature
                        }
                    }
                }"""})
    return self.errorhandler(response, "uploadDataset")


def getDatasetRuns(self, workspaceId, datasetId, state=None):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "getDatasetRuns",
            "variables": {
                "workspaceId": workspaceId,
                "datasetId": datasetId,
                "state": state,
            },
            "query": """query 
                getDatasetRuns($workspaceId: String!, $datasetId: String!, $state: String) {
                    getDatasetRuns(workspaceId: $workspaceId, datasetId: $datasetId, state: $state) {
                        runId
                        datasetId
                        workspaceId
                        channelId
                        startTime
                        endTime
                        state
                        run
                    }
                }"""})
    return self.errorhandler(response, "getDatasetRuns")


def getDatasetLog(self, workspaceId, datasetId, runId):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "getDatasetLog",
            "variables": {
                "workspaceId": workspaceId,
                "datasetId": datasetId,
                "runId": runId
            },
            "query": """query 
                getDatasetLog($workspaceId: String!, $datasetId: String!, $runId: String!) {
                    getDatasetLog(workspaceId: $workspaceId, datasetId: $datasetId, runId: $runId) {
                        runId
                        datasetId
                        workspaceId
                        channelId
                        startTime
                        endTime
                        state
                        email
                        run
                        log
                    }
                }"""})
    return self.errorhandler(response, "getDatasetLog")


# Copyright 2019-2022 DADoES, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License in the root directory in the "LICENSE" file or at:
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from abc import ABC
import logging
from anatools.lib.ana_object import AnaObject
import anatools.lib.context as ctx

logger = logging.getLogger(__name__)

class RiggedObject(AnaObject):

    def modify_rig(self):
        #Apply the modifiers in the dictionary to the specified objects

        if 'RigModifiers' not in self.config:
            print("WARNING: Rigged object does not have a modification list")
            return
        
        for modification in self.config["RigModifiers"]:

            Armature_name = modification['armature_name']
            Bone_name = modification['bone_name']

            Arm =  self.find_object([Armature_name])#First lets find the object we want to modifiy

            pose_bone = Arm.pose.bones[Bone_name]
            
            if modification['type']=='rotate':
                min = modification['min']
                max = modification['max']
                rot_axis = modification['axis']
                angle = min + ctx.random.random()*(max-min)
                angle_axis = [angle]
                angle_axis.extend(rot_axis)

                print("Rotating with", angle_axis)                
                pose_bone.rotation_mode = 'AXIS_ANGLE'
                pose_bone.rotation_axis_angle = angle_axis

                applied_as = {"Rotation (angle, axis)": angle_axis}

                self.modifiers.append({
                    "rig_modifier": {
                        "modifier": modification,
                        "rig_modifier_applied_as": applied_as
                    }  
                })

            if modification['type']=='translate':
                min = modification['min']
                max = modification['max']
                loc_axis = modification['axis']
                translation = min + ctx.random.random()*(max-min)
                translation_axis = [translation]
                translation_axis.extend(loc_axis)

                print("Translating ", translation)                
                pose_bone.location = (
                    loc_axis[0] * translation,
                    loc_axis[1] * translation,
                    loc_axis[2] * translation)

                applied_as = {"Translation (translation, axis)": translation_axis}

                self.modifiers.append({
                    "rig_modifier": {
                        "modifier": modification,
                        "rig_modifier_applied_as": applied_as
                    }  
                })

            else:
                print("WARNING: Unrecognized rig modification")




"""
Organizations API calls.
"""

def getOrganizations(self, organizationId=None):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "getOrganizations",
            "variables": {
                "organizationId": organizationId
            },
            "query": """query 
                getOrganizations($organizationId: String) {
                    getOrganizations(organizationId: $organizationId) {
                        organizationId
                        name
                        role
                    }
                }"""})
    return self.errorhandler(response, "getOrganizations")


def editOrganization(self, organizationId, name):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "editOrganization",
            "variables": {
                "organizationId": organizationId,
                "name": name
            },
            "query": """mutation 
                editOrganization($organizationId: String!, $name: String!) {
                    editOrganization(organizationId: $organizationId, name: $name) 
                }"""})
    return self.errorhandler(response, "editOrganization")



# Copyright 2019-2022 DADoES, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License in the root directory in the "LICENSE" file or at:
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import time
from numpy.random import RandomState

initialized = False
channel = None
seed = None
interp_num = None
preview = None
output = None
data = None
random = None
packages = None
channel = None

def initialize(channel, seed=None, interp_num=0, preview=False, output="./output", data="./data"):
    """ Initialize Ana configuration """

    globals()['channel'] = channel

    if seed is None:
        globals()['seed'] = int(str(int(time.time()*1e7))[-9:])
    else:
        globals()['seed'] = seed

    globals()['interp_num'] = interp_num
    globals()['preview'] = preview

    # create output directory if it doesn't already exist
    if output is not None:
        if not os.path.exists(output):
            os.makedirs(output)
    globals()['output'] = output

    # set the data directory
    globals()['data'] = data

    # use this for repeatable random distributions, e.g. self.ctx.ana_random.uniform(0,1)
    globals()['random'] = RandomState(globals()['seed'] + interp_num)

    # make package configurations available to all nodes
    globals()['packages'] = channel.packages

    globals()['initialized'] = True
    


"""
Workspace Functions
"""
    
def get_workspace(self):
    """Get Workspace ID of current workspace. 
    
    Returns
    -------
    str
        Workspace ID of current workspace.
    """
    if self.check_logout(): return
    return self.workspace


def set_workspace(self, workspaceId):
    """Set the workspace to the one you wish to work in.
    
    Parameters
    ----------
    workspaceId : str
        Workspace ID for the workspace you wish to work in.
    """
    if self.check_logout(): return
    if workspaceId is None: raise Exception('WorkspaceId must be specified.')
    workspaceSet = False
    self.workspaces = self.ana_api.getWorkspaces()
    for workspace in self.workspaces:
        if workspaceId == workspace['workspaceId']:
            self.workspace = workspace['workspaceId']
            self.organization = workspace['organizationId']
            workspaceSet = True
            break
    if not workspaceSet: raise Exception('Could not find workspace specified.')
    print(f'Organization set to {self.organization}.')
    print(f'Workspace set to {self.workspace}.')
    return


def get_workspaces(self, organizationId=None, workspaceId=None):
    """Shows list of workspaces with id, name, and owner data.
    
    Parameters
    ----------
    organizationId : str
        Organization ID to filter on. Optional
    workspaceId : str
        Workspace ID to filter on. Optional

    Returns
    -------
    list[dict]
        Workspace data for all workspaces for a user.
    """  
    if self.check_logout(): return
    if organizationId is None and workspaceId is None:
        self.workspaces = self.ana_api.getWorkspaces(organizationId, workspaceId)
        return self.workspaces
    else:
        workspaces = self.ana_api.getWorkspaces(organizationId, workspaceId)
        return workspaces


def create_workspace(self, name, channelIds=[], volumeIds=[], code=None):
    """Create a new workspace with specific channels.
    
    Parameters
    ----------
    name : str    
        New workspace name.
    channelIds : list[str]
        List of channel ids to add to workspace.
    volumeIds: list[str]
        List of volume ids that the workspace will have access to.
    code: str
        Content code that used for creating a workspace
    
    Returns
    -------
    str
        Workspace ID if creation was successful. Otherwise returns message.
    """    
    if self.check_logout(): return
    if name is None: raise ValueError("Name must be provided")
    if code is None: code = ''
    return self.ana_api.createWorkspace(organizationId=self.organization, name=name, channelIds = channelIds, volumeIds = volumeIds, code=code)


def delete_workspace(self, workspaceId=None, prompt=True):
    """Delete an existing workspace. 
    
    Parameters
    ----------
    workspaceId : str    
        Workspace ID for workspace to get deleted. Deletes current workspace if not specified. 
    prompt: bool
        Set to True if avoiding prompts for deleting workspace.
    
    Returns
    -------
    str
        Success or failure message if workspace was sucessfully removed.
    """
    if self.check_logout(): return
    if workspaceId is None: workspaceId = self.workspace 
    if prompt:
        response = input('This will remove any configurations, graphs and datasets associated with this workspace.\nAre you certain you want to delete this workspace? (y/n)  ')
        if response not in ['Y', 'y', 'Yes', 'yes']: return
    return self.ana_api.deleteWorkspace(workspaceId=workspaceId)


def edit_workspace(self, name=None, channelIds=None, volumeIds=None, workspaceId=None):
    """Edit workspace information. Provided channels list will result in the workspace having those channels. If channels is not provided, then no change will occur. 
    
    Parameters
    ----------
    name : str    
        New name to replace old one.
    channelIds: list[str]
        Names of channels that the workspace will have access to.
    volumeIds: list[str]
        List of volume ids that the workspace will have access to.
    workspaceId : str    
        Workspace ID for workspace to update.
    
    Returns
    -------
    str
        Success or failure message if workspace was sucessfully updated.
    """  
    if self.check_logout(): return
    if name is None and channelIds is None and volumeIds is None: return
    if workspaceId is None: workspaceId = self.workspace
    return self.ana_api.editWorkspace(workspaceId=workspaceId, name=name, channelIds=channelIds, volumeIds=volumeIds)


def get_workspace_guests(self, workspaceId=None):
    """Get guests of a workspace. Uses default workspace if not specified.
    
    Parameters
    ----------
    workspaceId : str
        Workspace Id. Optional.
    
    Returns
    -------
    list[dict]
        Information about guests of an workspace.
    """
    if self.check_logout(): return
    if workspaceId is None: workspaceId = self.workspace
    return self.ana_api.getMembers(workspaceId=workspaceId)

def get_workspace_invites(self, workspaceId=None):
    """Get invitations of a workspace. Uses default workspace if not specified.
    
    Parameters
    ----------
    workspaceId : str
        Workspace Id. Optional.
    
    Returns
    -------
    list[dict]
        Information about invitations of an workspace.
    """
    if self.check_logout(): return
    if workspaceId is None: workspaceId = self.workspace
    return self.ana_api.getInvitations(workspaceId=workspaceId)


def add_workspace_guest(self, email, workspaceId=None):
    """Add a guest to an existing workspace.
    
    Parameters
    ----------
    email: str
        Email of guest to add.
    workspaceId : str    
        Workspace ID to add a guest to. Uses current if not specified.
    
    Returns
    -------
    str
        Response status if guest got added to workspace succesfully. 
    """
    if self.check_logout(): return
    if email is None: raise ValueError("Email must be provided.")
    if workspaceId is None: workspaceId = self.workspace
    return self.ana_api.addMember(email=email, role=None, organizationId=None, workspaceId=workspaceId)


def remove_workspace_guest(self, email, workspaceId=None):
    """Remove a guest from an existing workspace.
    
    Parameters
    ----------
    email : str
        Guest email to remove.
    workspaceId : str    
        Workspace ID to remove guest from. Removes from current workspace if not specified. 
    
    Returns
    -------
    str
        Response status if guest got removed from workspace succesfully. 
    """
    if self.check_logout(): return
    if email is None: raise ValueError("Email must be provided.")
    if workspaceId is None: workspaceId = self.workspace
    return self.ana_api.removeMember(email=email, organizationId=None, workspaceId=workspaceId)

def remove_workspace_invitation(self, email, workspaceId=None, invitationId=None ):
    """Remove a invitation from an existing organization.
    
    Parameters
    ----------
    email : str
        Invitation email to remove.
    workspaceId: str
        Workspace ID to remove member from. Removes from current organization if not specified.
    inviteId: str
        Invitation ID to remove invitation from. Removes from current organization if not specified.
    
    Returns
    -------
    str
        Response status if member got removed from organization succesfully. 
    """
    if self.check_logout(): return
    if email is None: raise ValueError("Email must be provided.")
    if invitationId is None: raise ValueError("No invitation found.")
    if workspaceId is None: workspaceId = self.organization
    return self.ana_api.removeMember(email=email, workspaceId=workspaceId, organizationId=None, invitationId=invitationId)


def get_workspace_limits(self, setting=None, workspaceId=None):
    """Get information about Workspace limits and settings.

    Parameters
    ----------
    setting : str
        Setting name.
    workspaceId : str
        Workspace ID. Defaults to current if not specified.
    
    Returns
    -------
    list[dict]
        Workspace limit information.
    """
    if self.check_logout(): return
    if workspaceId is None: workspaceId = self.workspace
    return self.ana_api.getWorkspaceLimits(workspaceId=workspaceId, setting=setting)


def set_workspace_limit(self, setting, limit, workspaceId=None):
    """Set a limit for a workspace.

    Parameters
    ----------
    setting : str
        Setting name.
    limit : int
        Limit to set at.
    workspaceId : str
        Workspace ID. Defaults to current if not specified.
    
    Returns
    -------
    list[dict]
        Workspace limit information.
    """
    if self.check_logout(): return
    if setting is None: raise("Setting must be provided.")
    if limit is None: raise("Limit must be provided.")
    if workspaceId is None: workspaceId = self.workspace
    return self.ana_api.setWorkspaceLimit(workspaceId=workspaceId, setting=setting, limit=limit)




"""
Workspaces API calls.
"""

def getWorkspaces(self, organizationId=None, workspaceId=None):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "getWorkspaces",
            "variables": {
                "organizationId": organizationId,
                "workspaceId": workspaceId
            },
            "query": """query 
                getWorkspaces($organizationId: String, $workspaceId: String) {
                    getWorkspaces(organizationId: $organizationId, workspaceId: $workspaceId) {
                        workspaceId
                        organizationId
                        name
                        createdBy
                        createdAt
                        updatedAt
                    }
                }"""})
    return self.errorhandler(response, "getWorkspaces")


def createWorkspace(self, organizationId, name, channelIds, volumeIds, code):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "createWorkspace",
            "variables": {
                "organizationId": organizationId,
                "name": name,
                "channelIds": channelIds,
                "volumeIds": volumeIds,
                "code": code
            },
            "query": """mutation 
                createWorkspace($organizationId: String!, $name: String!, $channelIds: [String]!, $volumeIds: [String]!, $code: String!) {
                    createWorkspace(organizationId: $organizationId, name: $name, channelIds: $channelIds, volumeIds: $volumeIds, code: $code)
                }"""})
    return self.errorhandler(response, "createWorkspace")


def deleteWorkspace(self, workspaceId):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "deleteWorkspace",
            "variables": {
                "workspaceId": workspaceId
            },
            "query": """mutation 
                deleteWorkspace($workspaceId: String!) {
                    deleteWorkspace(workspaceId: $workspaceId)
                }"""})
    return self.errorhandler(response, "deleteWorkspace")


def editWorkspace(self, workspaceId, name=None, channelIds=None, volumeIds=None):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "editWorkspace",
            "variables": {
                "workspaceId": workspaceId,
                "name": name,
                "channelIds": channelIds,
                "volumeIds": volumeIds,
            },
            "query": """mutation 
                editWorkspace($workspaceId: String!, $name: String, $channelIds: [String], $volumeIds: [String]) {
                    editWorkspace(workspaceId: $workspaceId, name: $name, channelIds: $channelIds, volumeIds: $volumeIds)
                }"""})
    return self.errorhandler(response, "editWorkspace")



"""
GAN API calls.
"""

def getGANModels(self, workspaceId, modelId):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "getGANModels",
            "variables": {
                "workspaceId": workspaceId,
                "modelId": modelId
            },
            "query": """query 
                getGANModels($workspaceId: String!, $modelId: String) {
                    getGANModels(workspaceId: $workspaceId, modelId: $modelId){
                        modelId
                        name
                        description
                    }
                }"""})
    return self.errorhandler(response, "getGANModels")


def getGANDataset(self, datasetId, workspaceId):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "getGANDataset",
            "variables": {
                "workspaceId": workspaceId,
                "datasetId": datasetId
            },
            "query": """query 
                getGANDataset($workspaceId: String!, $datasetId: String!) {
                    getGANDataset(workspaceId: $workspaceId, datasetId: $datasetId) {
                        datasetId: datasetid
                        channelId
                        graphId: source
                        parentDataset: ganparent
                        modelId: ganmodelId
                        interpretations: scenarios
                        user
                        status
                        files
                        size
                        name
                        description
                    }
                }"""})
    return self.errorhandler(response, "getGANDataset")


def createGANDataset(self, modelId, datasetId, workspaceId):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "createGANDataset",
            "variables": {
                "workspaceId": workspaceId,
                "datasetId": datasetId,
                "modelId": modelId,
            },
            "query": """mutation 
                createGANDataset($workspaceId: String!, $datasetId: String!, $modelId: String!) {
                    createGANDataset(workspaceId: $workspaceId, datasetId: $datasetId, modelId: $modelId)
                }"""})
    return self.errorhandler(response, "createGANDataset")


def deleteGANDataset(self, datasetId, workspaceId):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "deleteGANDataset",
            "variables": {
                "workspaceId": workspaceId,
                "datasetId": datasetId
            },
            "query": """mutation 
                deleteGANDataset($workspaceId: String!, $datasetId: String!) {
                    deleteGANDataset(workspaceId: $workspaceId, datasetId: $datasetId)
                }"""})
    return self.errorhandler(response, "deleteGANDataset")


def uploadGANModel(self, organizationId, name, description, flags):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "uploadGANModel",
            "variables": {
                "organizationId": organizationId,
                "name": name,
                "description": description,
                "flags": flags,
            },
            "query": """mutation 
                uploadGANModel($organizationId: String!, $name: String!, $description: String!, $flags: String) {
                    uploadGANModel(organizationId: $organizationId, name: $name, description: $description, flags: $flags){
                        key
                        modelId
                        url
                        fields {
                            key
                            bucket
                            algorithm
                            credential
                            date
                            token
                            policy
                            signature
                        }
                    }
                }"""})
    return self.errorhandler(response, "uploadGANModel")


def deleteGANModel(self, modelId):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "deleteGANModel",
            "variables": {
                "modelId": modelId,
            },
            "query": """mutation 
                deleteGANModel($modelId: String!) {
                    deleteGANModel(modelId: $modelId)
                }"""})
    return self.errorhandler(response, "deleteGANModel")


def addGANAccess(self, organizationId, modelId):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "addGANAccess",
            "variables": {
                "organizationId": organizationId,
                "modelId": modelId,
            },
            "query": """mutation 
                addGANAccess($organizationId: String!, $modelId: String!) {
                    addGANAccess(organizationId: $organizationId, modelId: $modelId)
                }"""})
    return self.errorhandler(response, "addGANAccess")


def removeGANAccess(self, organizationId, modelId):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "removeGANAccess",
            "variables": {
                "organizationId": organizationId,
                "modelId": modelId,
            },
            "query": """mutation 
                removeGANAccess($organizationId: String!, $modelId: String!) {
                    removeGANAccess(organizationId: $organizationId, modelId: $modelId)
                }"""})
    return self.errorhandler(response, "removeGANAccess")


# Copyright 2019-2022 DADoES, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License in the root directory in the "LICENSE" file or at:
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from cmath import log
import os
import logging
import anatools.lib.context as ctx

logger = logging.getLogger(__name__)

def get_volume_path(package, inpath):
    """
    Convert a volume path to an absolute path
    """
    if ":" in inpath:
        # path includes a volume
        volume_name, rel_path = inpath.split(":")
        volume_path = ctx.packages[package]['volumes'][volume_name]
        if os.path.isabs(volume_path):
            # volume is an absolute path
            return os.path.join(volume_path, rel_path)
        else:
            # volume is relative to "--data" parameter
            if volume_path == 'local':
                return os.path.join(ctx.data, 'local', volume_name, rel_path)
            else: 
                return os.path.join(ctx.data, 'volumes', volume_path, rel_path)
    else:
        # path does not include a volume
        if os.path.isabs(inpath):
            # path is absolute
            return inpath
        else:
            # path is relative to "--data" parameter
            return os.path.join(ctx.data, inpath)



import numpy as np
from PIL import Image, ImageDraw
import json
import os


def draw(image_path, out_dir, draw_type='box_2d', object_ids=None, object_types=None, line_thickness=1): 
    """
    This function handles the io and draws the right type of annotation on the image.

    Parameters
    ----------
    image_path : str
        Path to of specific image file to draw the boxes for.  
    out_dir : str
        File path to directory where the image should be saved to.
    draw_type : str
        Draw either a 2d bounding box, 3d bounding box, or segmentation on objects within an image. Must pass in either 'box_2d', 'box_3d', or 'segmentation' for values.
    object_ids : list[int]
        List of object id's to annotate. If not provided, all objects will get annotated. Choose either id or type filter.
    object_types: list[str]
        Filter for the object types to annotate. If not provided, all object types will get annotated. Choose either id or type filter.
    line_thickness: int
        Desired line thickness for box outline. 
    """

    if not os.path.exists(image_path):
        print('Incorrect path to images: ' + image_path)
        return

    if object_types and object_ids:
        print('Provide either object_types OR object_ids. ')
        return

    if not os.path.exists(out_dir):
        os.makedirs(out_dir)

    root_dir = ('/').join(image_path.split('/')[:-2])
    image_name = image_path.split('/')[-1].split('.')[0]
    image_ext = image_path.split('/')[-1].split('.')[1]

    annotation_file = root_dir+'/annotations/'+image_name+"-ana.json"
    file = open(annotation_file)
    annotations = json.load(file)
    file.close()
    
    annotation_ids = [data['id'] for data in annotations['annotations']]
    if object_ids is not None and not check_lists(annotation_ids, object_ids, 'object_ids'):
        return
    
    
    metadata_file = root_dir+'/metadata/'+image_name+'-metadata.json'
    file = open(metadata_file)
    metadata = json.load(file)
    file.close()                    

    metadata_types = list(set([data['type'] for data in metadata['objects']]))
    
    # get object id and type to correspond to a unique color per type
    if object_ids:
        object_data = [data for data in metadata['objects'] if data['id'] in object_ids]
    elif object_types is not None:
        if not check_lists(metadata_types, object_types, 'object_types'):
            return  
        object_data = [data for data in metadata['objects'] if data['type'] in object_types]
    else:
        object_data = [data for data in metadata['objects']]

    # generate a unique color for each object type
    type_colors = {}

    for type in metadata_types:
        generate_color = [int(val) for val in np.random.randint(0, 255, 3)]
        type_colors[type] = tuple(generate_color)

    draw_img = Image.open(image_path)
    for object in annotations['annotations']:
        if any(object['id'] == d['id'] for d in object_data):
            color_info = [data for data in object_data if object['id'] == data['id']]
            if draw_type == 'box_2d':
                draw_img = box_2d(object['bbox'], draw_img, line_thickness, type_colors[color_info[0]['type']])
            elif draw_type == 'box_3d' :
                draw_img = box_3d(object['bbox3d'], draw_img, line_thickness, type_colors[color_info[0]['type']])
            elif draw_type == 'segmentation':
                draw_img = segmentation(object['segmentation'], draw_img, line_thickness, type_colors[color_info[0]['type']])
            else:
                print('Provide either box_2d, box_3d, or segmentation')

    outimg = out_dir+'/'+image_name+'-annotated-'+draw_type+'.'+image_ext
    draw_img.save(outimg)
    print(f'Image saved to {outimg}')


def box_2d(coordinates, bbox_img, width, outline):
    """
    Draws 2d boxes around objects given a list of coordinates.

    Parameters
    ----------
    coordinates : list
        The list of coordinates.
    bbox_img : array
        Image in form of a numpy array.
    width: int
        Desired line thickness for box outline. 
    outline: list
        List of bgr color values for PIL color inputs.

    Returns
    -------
    array
        Image in form of a numpy array with 2d boxes around objects.
    """
    x,y,w,h = coordinates
    draw = ImageDraw.Draw(bbox_img)
    draw.rectangle(((x,y), (x+w, y+h)), None, outline, width)
    return bbox_img


def box_3d(coordinates, bbox_img, width, fill):
    """
    Draws 3d boxes around objects given a list of coordinates.

    Parameters
    ----------
    coordinates : list
        The list of coordinates.
    bbox_img : array
        Image in form of a numpy array.
    width: int
        Desired line thickness for box outline. 
    fill: list
        List of bgr color values for PIL color inputs.
    
    Returns
    -------
    array
        Image in form of a numpy array with 3d boxes around objects.
    """
    point_1 = (coordinates[0], coordinates[1])
    point_2 = (coordinates[3], coordinates[4])
    point_3 = (coordinates[6], coordinates[7])
    point_4 = (coordinates[9], coordinates[10])
    point_5 = (coordinates[12], coordinates[13])
    point_6 = (coordinates[15], coordinates[16])
    point_7 = (coordinates[18], coordinates[19])
    point_8 = (coordinates[21], coordinates[22])
    draw = ImageDraw.Draw(bbox_img)
    draw.line((point_1, point_2), fill, width)
    draw.line((point_1, point_5), fill, width)
    draw.line((point_5, point_6), fill, width)
    draw.line((point_2, point_6), fill, width)
    draw.line((point_3, point_4), fill, width)
    draw.line((point_3, point_7), fill, width)
    draw.line((point_7, point_8), fill, width)
    draw.line((point_4, point_8), fill, width)
    draw.line((point_3, point_2), fill, width)
    draw.line((point_7, point_6), fill, width)
    draw.line((point_4, point_1), fill, width)
    draw.line((point_8, point_5), fill, width)
    return bbox_img


def segmentation(coordinates, draw_img, width, color):
    """
    Draws an outline around objects given a list of coordinates.

    Parameters
    ----------
    coordinates : list
        The list of coordinates.
    draw_img : array
        Image in form of a numpy array.
    width: int
        Desired line thickness for box outline. 
    fill: list
        List of bgr color values for cv2 color inputs.

    Returns
    -------
    array
        Image in form of a numpy array with objects outlined.
    """
    draw = ImageDraw.Draw(draw_img)
    for poly in coordinates:
        draw.polygon(poly, fill=None, outline=color, width=width)
    return draw_img


def check_lists(actual, expected, name_to_check):
    """
    Helper function that checks if one list is in another for validation on draw inputs.

    Parameters
    ----------
    actual : list
        The list of actual values found in either annotation or metadata file.
    expected : list
        The expected list that is provided from the user.
    name_to_check: str
        Name of parameter to check (either object_id or object_type).

    Returns
    -------
    bool
        True if lists are matching, False otherwise.
    """
    
    if not set(expected).issubset(set(actual)):
        out_of_bounds_check= list(set(expected) - set(actual))
        print(f'Provided {name_to_check} list has the following out of bounds: {out_of_bounds_check}. Please rerun with valid list. \nHere are all the {name_to_check} that can get annotated: ')
        print(list(set(actual)))
        return False
    return True


import json
import yaml
import os
import datetime
from PIL import Image

def create_cocodata():
    cocodata = dict()
    cocodata['info'] = {    
        "description":  "Rendered.AI Synthetic Dataset",
        "url":          "https://rendered.ai/",
        "contributor":  "info@rendered.ai",
        "version":      "1.0",
        "year":         str(datetime.datetime.now().year),
        "date_created": datetime.datetime.now().isoformat()}
    cocodata['licenses'] = [{
        "id":   0,
        "url":  "https://rendered.ai/",     # "url": "https://creativecommons.org/licenses/by-nc-nd/4.0/",
        "name": "Rendered.AI License"}]     # "name": "Creative Commons Attribution-NonCommercial-NoDerivatives 4.0 International License"}]
    cocodata['images'] = list()
    cocodata['categories'] = list()
    cocodata['annotations'] = list()
    return cocodata


def convert_coco(datadir, outdir, mapping):

    annsdir = os.path.join(datadir, "annotations")
    metadir = os.path.join(datadir, "metadata")
    imgdir = os.path.join(datadir, "images")
    annsfiles = os.listdir(annsdir)
    
    cocodata = create_cocodata()
    annotations = []
    categories = []
    cats = []
    imgid = 0
    annid = 0
        
     # for each interpretation, gather annotations and map categories
    for f in sorted(annsfiles):
        if not f.endswith('.json'):
            continue
        with open(os.path.join(annsdir,f), 'r') as af: anns = json.load(af)
        with open(os.path.join(metadir,f.replace('ana','metadata')), 'r') as mf: metadata = json.load(mf)
        
        # for each object in the metadata file, check if any of the properties are true
        for obj in metadata['objects']:
            for prop in mapping['properties']:
                if eval(prop):
                    for ann in anns['annotations']:
                        if ann['id'] == obj['id']: 
                            cat = mapping['classes'][mapping['properties'][prop]]
                            if cat not in cats: cats.append(cat)
                            annotation = {}
                            annotation['id'] = annid
                            annotation['image_id'] = imgid
                            annotation['category_id'] = cats.index(cat)
                            annotation['segmentation'] = ann['segmentation']
                            annotation['area'] = ann['bbox'][2] * ann['bbox'][3]
                            annotation['bbox'] = ann['bbox']
                            annotation['iscrowd'] = 0
                            annid += 1
                            cocodata['annotations'].append(annotation)
                            break
        imgdata = {
            'id':               imgid, 
            'file_name':        metadata['filename'], 
            'date_captured':    metadata['date'], 
            'license':          0 }
        if 'sensor' in metadata:
            metadata['width'] =  metadata['sensor']['resolution'][0],
            metadata['height']=  metadata['sensor']['resolution'][1],
            if 'frame' in metadata['sensor']: metadata['frame'] = metadata['sensor']['frame']
        else:
            im = Image.open(os.path.join(imgdir, anns['filename']))
            width, height = im.size
            metadata['width'] =  width
            metadata['height']=  height
        cocodata['images'].append(imgdata)
        imgid += 1
    for cat in cats:
        cocodata['categories'].append({
            'id':               cats.index(cat), 
            'name':             cat[-1],
            'supercategory':    cat[0]
        })

    with open(os.path.join(outdir,'coco.json'), 'w+') as of:
        json.dump(cocodata,of)



"""
GAN API calls.
"""

def getUMAP(self, umapId, workspaceId):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "getUMAP",
            "variables": {
                "workspaceId": workspaceId,
                "umapId": umapId
            },
            "query": """query 
                getUMAP($workspaceId: String!, $umapId: String) {
                    getUMAP(workspaceId: $workspaceId, umapId: $umapId){
                        datasets
                        samples
                        seed
                        status
                        results {
                            id
                            featuresize
                            datasets {
                                datasetId
                                datasetName
                                points {
                                    image
                                    x
                                    y
                                    z
                                }
                            }
                        }
                    }
                }"""})
    return self.errorhandler(response, "getUMAP")


def createUMAP(self, datasetIds, samples, workspaceId):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "createUMAP",
            "variables": {
                "workspaceId": workspaceId,
                "datasetIds": datasetIds,
                "samples": samples
            },
            "query": """query 
                createUMAP($workspaceId: String!, $datasetIds: [String]!, $samples: [Int]!) {
                    createUMAP(workspaceId: $workspaceId, datasetIds: $datasetIds, samples: $samples)
                }"""})
    return self.errorhandler(response, "createUMAP")


def deleteUMAP(self, umapId, workspaceId):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "deleteUMAP",
            "variables": {
                "workspaceId": workspaceId,
                "umapId": umapId,
            },
            "query": """mutation 
                deleteUMAP($workspaceId: String!, $umapId: String!) {
                    deleteUMAP(workspaceId: $workspaceId, umapId: $umapId)
                }"""})
    return self.errorhandler(response, "deleteUMAP")




# Copyright 2019-2022 DADoES, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License in the root directory in the "LICENSE" file or at:
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
def camera_as_planes(scene, obj):
    """
    Return planes in world-space which represent the camera view bounds.
    """
    from mathutils.geometry import normal

    camera = obj.data
    # normalize to ignore camera scale
    matrix = obj.matrix_world.normalized()
    frame = [matrix @ v for v in camera.view_frame(scene=scene)]
    origin = matrix.to_translation()

    planes = []
    from mathutils import Vector
    is_persp = (camera.type != 'ORTHO')
    for i in range(4):
        # find the 3rd point to define the planes direction
        if is_persp:
            frame_other = origin
        else:
            frame_other = frame[i] + matrix.col[2].xyz

        n = normal(frame_other, frame[i - 1], frame[i])
        d = -n.dot(frame_other)
        planes.append((n, d))

    if not is_persp:
        # add a 5th plane to ignore objects behind the view
        n = normal(frame[0], frame[1], frame[2])
        d = -n.dot(origin)
        planes.append((n, d))

    return planes


def side_of_plane(p, v):
    return p[0].dot(v) + p[1]


def is_segment_in_planes(p1, p2, planes):
    dp = p2 - p1

    p1_fac = 0.0
    p2_fac = 1.0

    for p in planes:
        div = dp.dot(p[0])
        if div != 0.0:
            t = -side_of_plane(p, p1)
            if div > 0.0:
                # clip p1 lower bounds
                if t >= div:
                    return False
                if t > 0.0:
                    fac = (t / div)
                    p1_fac = max(fac, p1_fac)
                    if p1_fac > p2_fac:
                        return False
            elif div < 0.0:
                # clip p2 upper bounds
                if t > 0.0:
                    return False
                if t > div:
                    fac = (t / div)
                    p2_fac = min(fac, p2_fac)
                    if p1_fac > p2_fac:
                        return False

    ## If we want the points
    # p1_clip = p1.lerp(p2, p1_fac)
    # p2_clip = p1.lerp(p2, p2_fac)        
    return True


def point_in_object(obj, pt):
    xs = [v[0] for v in obj.bound_box]
    ys = [v[1] for v in obj.bound_box]
    zs = [v[2] for v in obj.bound_box]
    pt = obj.matrix_world.inverted() @ pt
    return (min(xs) <= pt.x <= max(xs) and
            min(ys) <= pt.y <= max(ys) and
            min(zs) <= pt.z <= max(zs))


def object_in_planes(obj, planes):
    from mathutils import Vector

    matrix = obj.matrix_world
    box = [matrix @ Vector(v) for v in obj.bound_box]
    for v in box:
        if all(side_of_plane(p, v) > 0.0 for p in planes):
            # one point was in all planes
            return True

    # possible one of our edges intersects
    edges = ((0, 1), (0, 3), (0, 4), (1, 2),
             (1, 5), (2, 3), (2, 6), (3, 7),
             (4, 5), (4, 7), (5, 6), (6, 7))
    if any(is_segment_in_planes(box[e[0]], box[e[1]], planes)
           for e in edges):
        return True


    return False


def objects_in_planes(objects, planes, origin):
    """
    Return all objects which are inside (even partially) all planes.
    """
    return [obj for obj in objects
            if point_in_object(obj, origin) or
               object_in_planes(obj, planes)]

def objects_in_camera(objects,camera):
    from bpy import context
    scene = context.scene
    origin = camera.matrix_world.to_translation()
    planes = camera_as_planes(scene, camera)
    objects_in_view = []
    for obj in objects:
        meshes_objects = collect_mesh_objects(obj)
        if len(objects_in_planes(meshes_objects, planes, origin)) is not 0:
            objects_in_view.append(obj) #Some part of this object is visible

    return objects_in_view

def collect_mesh_objects(obj):
    object_array=[]
    if obj.type=='MESH': object_array.append(obj)
    # print("checking", obj.name)
    # print(len(obj.children), "children found.")
    for child in obj.children:
        collection = collect_mesh_objects(child)
        if len(collection) is not 0:
            for item in collection:
                object_array.append(item)
    return object_array

#    for obj in objects_in_view:
#        obj.select_set(True)
        
if __name__ == "__main__":
    import bpy
    scene = bpy.context.scene
    print(len(objects_in_camera([bpy.data.objects['Remy']], scene.camera)))



"""
Annotations Functions
"""

def get_annotation_formats(self):
    """Retrieves the annotation formats supported by the Platform.
    
    Returns
    -------
    str
        The annotation formats supported by the Platform.
    """
    if self.check_logout(): return
    return self.ana_api.getAnnotationFormats()


def get_annotation_maps(self, channelId):
    """Retrieves the map files that are supported by the channel.
    
    Parameters
    ----------
    channelId : str
        Channel ID to retrieve maps for.
    
    Returns
    -------
    str
        The annotation maps supported by the channel.
    """
    if self.check_logout(): return
    if channelId is None: raise ValueError('ChannelId must be specified.')
    return self.ana_api.getAnnotationMaps(channelId)


def get_annotations(self, datasetId=None, annotationId=None, workspaceId=None):
    """Retrieve information about existing annotations generated for a dataset. Querying requires both datasetId and annotationId.
    
    Parameters
    ----------
    datasetId : str
        Dataset ID to generate annotations for.
    annotationId : str
        Annotation ID for a specific annotations job.
    workspaceId: str
        Workspace ID where the annotations exist. If none is provided, the current workspace will get used. 
    
    Returns
    -------
    list[dict]
        Annotation information.
    """
    if self.check_logout(): return
    if annotationId is None and datasetId is None: raise ValueError('datasetId and annotationId must be specified.')
    if workspaceId is None: workspaceId = self.workspace
    return self.ana_api.getAnnotations(workspaceId=workspaceId, datasetId=datasetId, annotationId=annotationId)
    

def create_annotation(self, datasetId, format, map, workspaceId=None):
    """Generates annotations for an existing dataset. 
    
    Parameters
    ----------
    datasetId : str
        Dataset ID to generate annotation for.
    format : str
        Annotation format. Call get_annotation_formats() to find supported formats.
    map: str
        The map file used for annotations. Call get_annotation_maps() to find supported maps.
    workspaceId: str
        Workspace ID of the dataset to generate annotation for. If none is provided, the current workspace will get used. 
    
    Returns
    -------
    str
        The annotationsId for the annotation job.
    """
    if self.check_logout(): return
    if datasetId is None: raise ValueError("DatasetId must be defined.")
    if format is None: raise ValueError("Format must be defined.")
    if map is None: raise ValueError("Map must be defined.")
    if workspaceId is None: workspaceId = self.workspace
    return self.ana_api.createAnnotation(workspaceId=workspaceId, datasetId=datasetId, format=format, map=map)
    

def download_annotation(self, datasetId, annotationId, workspaceId=None):
    """Downloads annotations archive.
    
    Parameters
    ----------
    datasetId : str
        Dataset ID to download image annotation for.
    annotationId : str
        Id of previously generated image annotation. 
    workspaceId: str
        Workspace ID of the dataset to generate annotation for. If none is provided, the current workspace will get used. 
    
    Returns
    -------
    str
        The name of the archive file that got downloaded.
    """
    import requests
    if self.check_logout(): return
    if workspaceId is None: workspaceId = self.workspace
    url = self.ana_api.downloadAnnotation(workspaceId=workspaceId, datasetId=datasetId, annotationId=annotationId)
    fname = url.split('?')[0].split('/')[-1]
    with requests.get(url, stream=True) as downloadresponse:
        with open(fname, 'wb') as outfile:
            downloadresponse.raise_for_status()
            outfile.write(downloadresponse.content)
            with open(fname, 'wb') as f:
                for chunk in downloadresponse.iter_content(chunk_size=8192):
                    f.write(chunk)
    return fname


def delete_annotation(self, annotationId, workspaceId=None):
    """Delete a dataset annotation.
    
    Parameters
    ----------
    annotationId : str
        AnnoationId of the annotation job.
    workspaceId: str
        Workspace ID of the dataset to generate annotation for. If none is provided, the current workspace will get used. 
    
    Returns
    -------
    bool
        If true, successfully deleted the annotation.
    """
    if self.check_logout(): return
    if annotationId is None: raise ValueError("AnnotationId must be defined.")
    if workspaceId is None: workspaceId = self.workspace
    return self.ana_api.deleteAnnotation(workspaceId=workspaceId, annotationId=annotationId)




# Copyright 2019-2022 DADoES, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License in the root directory in the "LICENSE" file or at:
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import sys
import logging
import gc
import anatools.lib.context as ctx
from anatools.lib.node import create_node

logger = logging.getLogger(__name__)

def interp(graph):
    """Interpret a graph"""
    graphFormatVersion = graph.get("version", 0.0)

    # create and configure nodes
    nodes = {}
    for name, node_config in sorted(graph["nodes"].items()):
        if graphFormatVersion == 0.0:
            nodes[name] = create_node(name, node_config["class"])
            nodes[name].configure_v0(node_config)
        else:
            nodes[name] = create_node(name, node_config["nodeClass"])
            nodes[name].configure(node_config)

    # append outlinks to nodes
    for dst_node in nodes:
        for dst_port in nodes[dst_node].inlinks:
            for link in nodes[dst_node].inlinks[dst_port]:
                src_node, src_port = link
                if src_node in nodes.keys():
                    schema_output_set = set()
                    for port_dict in nodes[src_node].schema["outputs"]:
                        schema_output_set.add(port_dict["name"])
                    if src_port not in schema_output_set:
                        logger.error(
                            "Node '%s' class '%s' input '%s' is linked to node '%s' class '%s' on undefined port '%s'",
                            dst_node, nodes[dst_node].__class__.__name__, dst_port,
                            src_node, nodes[src_port].__class__.__name__, src_port)
                        sys.exit(1)
                    if src_port not in nodes[src_node].outlinks.keys():
                        nodes[src_node].outlinks[src_port] = []
                    nodes[src_node].outlinks[src_port].append((dst_node, dst_port))
                else:
                    logger.error(
                        "Node '%s' class '%s' input '%s' is linked to undefined node '%s'",
                        dst_node, nodes[dst_node].__class__.__name__, dst_port, src_node)
                    sys.exit(1)

    # execute nodes
    while len(nodes) > 0:
        node_executed = None
        for name, node in nodes.items():
            # if all input links have been resolved then execute this node
            if len(node.inlinks.keys()) == 0:
                logger.info("Executing node '%s' class '%s'", name, node.__class__.__name__)
                outputs = node.exec()
                # verify the actual node output matches the schema outputs
                # TODO: After people clean up their returns, make this throw an exception.
                schema_output_set = set()
                for port_dict in node.schema["outputs"]:
                    schema_output_set.add(port_dict["name"])
                if set(outputs.keys()) != schema_output_set:
                    logger.error("Output returned by node '%s' class '%s' doesn't match output defined in schema",
                                node.name, node.__class__.__name__)
                # resolve output links
                for src_port in node.outlinks:
                    for outlink in node.outlinks[src_port]:
                        dst_node, dst_port = outlink
                        # append value to input of destination node/port
                        if dst_port not in nodes[dst_node].inputs.keys():
                            nodes[dst_node].inputs[dst_port] = []
                        nodes[dst_node].inputs[dst_port].append(outputs[src_port])
                        # remove inlink from destination node/port
                        try:
                            nodes[dst_node].inlinks[dst_port].remove((name, src_port))
                            # if there are no more inlinks on the destination port then delete it
                            if len(nodes[dst_node].inlinks[dst_port]) == 0:
                                del nodes[dst_node].inlinks[dst_port]
                        except ValueError:
                            # this is a coding error
                            logger.critical(
                                "Node '%s' class '%s' is missing link '[%s, %s]'",
                                dst_node, nodes[dst_node].__class__.__name__, name, src_port)
                            sys.exit(1)
                node_executed = node
                break

        # delete the node we executed
        if node_executed:
            del nodes[node_executed.name]
        else:
            # this is either a coding error or a cycle in the graph
            errorString = "Graph execution failed; graph may be cyclic. The following links could not be resolved: "
            linkDescriptions = []
            for name in nodes:
                for port in nodes[name].inlinks:
                    for inlink in nodes[name].inlinks[port]:
                        linkDescriptions.append("Node '%s' class '%s' input '%s' link '[%s, %s]'",
                                                name, nodes[name].__class__.__name__, port, inlink[0], inlink[1])
            errorString += ', '.join(linkDescriptions)
            logger.error(errorString)
            sys.exit(1)

    gc.collect()


import json
import yaml
import os
from PIL import Image

def generate_xml(object, xml=None, level=0):
    if xml is None: xml = ''
    for key in object.keys():
        if (type(object[key]) is list):
            for val in object[key]:
                xml += '\t'*level+f'<object>\n'
                xml = generate_xml(val, xml, level+1)
                xml += '\t'*level+f'</object>\n'
        elif (type(object[key]) is dict): 
            xml += '\t'*level+f'<{key}>\n'
            xml = generate_xml(object[key], xml, level+1)
            xml += '\t'*level+f'</{key}>\n'
        else: xml += '\t'*level+f'<{key}>{object[key]}</{key}>\n'
    return xml


def convert_pascal(datadir, outdir, mapping):
    """ Generate annotations in PASCAL VOC format. Result will be placed in outdir.

    Parameters
    ----------
    datadir : str
        Location of Rendered.ai dataset output.
    outdir : str
        Location where the results should be written.
    mapfile: str
        The map file used for annotations (YAML only).
    
    Returns
    -------
    """
    annsdir = os.path.join(datadir, "annotations")
    metadir = os.path.join(datadir, "metadata")
    imagedir = os.path.join(datadir, "images")
    annsfiles = os.listdir(annsdir)

    # for each interpretation, gather annotations and map categories
    for f in sorted(annsfiles):
        with open(os.path.join(annsdir,f), 'r') as af: anns = json.load(af)
        with open(os.path.join(metadir,f.replace('ana','metadata')), 'r') as mf: metadata = json.load(mf)
        
        image = Image.open(os.path.join(imagedir,anns['filename']))
        width = image.size[0]
        height = image.size[1]
        depth = len(image.getbands())
        xmldata = {
            'annotation': {
                'folder': 'images',
                'filename': anns['filename'],
                'path': 'images/'+anns['filename'],
                'source': {'database': 'Unknown'},
                'size': {
                    'width': width,
                    'height': height,
                    'depth': depth
                },
                'segmented': 0,
                'objects': []
            }
        }

        # for each object in the metadata file, check if any of the properties are true
        for obj in metadata['objects']:
            for prop in mapping['properties']:
                if eval(prop):
                    for ann in anns['annotations']:
                        if ann['id'] == obj['id']: objann = ann
                    objectdata = {
                        'name': mapping['classes'][mapping['properties'][prop]][-1],
                        'pose': 'Unspecified',
                        'truncated': 0,
                        'difficult': 0,
                        'bndbox': {
                            'xmin': objann['bbox'][0],
                            'ymin': objann['bbox'][1],
                            'xmax': objann['bbox'][0] + objann['bbox'][2],
                            'ymax': objann['bbox'][1] + objann['bbox'][3]
                        }
                    }
                    xmldata['annotation']['objects'].append(objectdata)
                    break
        
        # write xmlfile
        text = generate_xml(xmldata)
        with open(os.path.join(outdir,f.replace('-ana.json','.xml')), 'w+') as xmlfile:
            xmlfile.write(text)


"""
Limits API calls.
"""

def getPlatformLimits(self, tier=None, setting=None):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "getPlatformLimits",
            "variables": {
                "tier": tier,
                "setting": setting
            },
            "query": """query 
                getPlatformLimits($tier: String, $setting: String) {
                    getPlatformLimits(tier: $tier, setting: $setting){
                        setting
                        tier
                        value
                    }
                }"""})
    return self.errorhandler(response, "getPlatformLimits")


def setPlatformLimit(self, tier, setting, limit):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "setPlatformLimits",
            "variables": {
                "setting": setting,
                "tier": tier,
                "limit": limit
            },
            "query": """mutation 
                setPlatformLimits($tier: String!, $setting: String!, $limit: Int!) {
                    setPlatformLimits(tier: $tier, setting: $setting, limit: $limit)
                }"""})
    return self.errorhandler(response, "setPlatformLimits")


def getOrganizationLimits(self, organizationId, setting=None):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "getOrganizationLimits",
            "variables": {
                "organizationId": organizationId,
                "setting": setting
            },
            "query": """query 
                getOrganizationLimits($organizationId: String!, $setting: String) {
                    getOrganizationLimits(organizationId: $organizationId, setting: $setting){
                        organizationId
                        setting
                        limit
                        usage
                    }
                }"""})
    return self.errorhandler(response, "getOrganizationLimits")

def setOrganizationLimit(self, organizationId, setting, limit):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "setOrganizationLimits",
            "variables": {
                "organizationId": organizationId,
                "setting": setting,
                "limit": limit
            },
            "query": """mutation 
                setOrganizationLimits($organizationId: String!, $setting: String!, $limit: Int!) {
                    setOrganizationLimits(organizationId: $organizationId, setting: $setting, limit: $limit)
                }"""})
    return self.errorhandler(response, "setOrganizationLimits")


def getWorkspaceLimits(self, workspaceId, setting=None):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "getWorkspaceLimits",
            "variables": {
                "workspaceId": workspaceId,
                "setting": setting
            },
            "query": """query 
                getWorkspaceLimits($workspaceId: String!, $setting: String) {
                    getWorkspaceLimits(workspaceId: $workspaceId, setting: $setting){
                        workspaceId
                        setting
                        limit
                        usage
                    }
                }"""})
    return self.errorhandler(response, "getWorkspaceLimits")


def setWorkspaceLimit(self, workspaceId, setting, limit):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "setWorkspaceLimits",
            "variables": {
                "workspaceId": workspaceId,
                "setting": setting,
                "limit": limit
            },
            "query": """mutation 
                setWorkspaceLimits($workspaceId: String!, $setting: String!, $limit: Int!) {
                    setWorkspaceLimits(workspaceId: $workspaceId, setting: $setting, limit: $limit)
                }"""})
    return self.errorhandler(response, "setWorkspaceLimits")


def getOrganizationUsage(self, organizationId, year, month, workspaceId=None, member=None):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "getOrganizationUsage",
            "variables": {
                "organizationId": organizationId,
                "workspaceId": workspaceId,
                "member": member,
                "year": year,
                "month": month
            },
            "query": """query 
                getOrganizationUsage($organizationId: String!, $workspaceId: String, $member: String, $year: String!, $month: String!) {
                    getOrganizationUsage(organizationId: $organizationId, workspaceId: $workspaceId, member: $member, year: $year, month: $month){
                        channel
                        instanceType
                        time
                    }
                }"""})
    return self.errorhandler(response, "getOrganizationUsage")




"""
Dataset Functions
"""

from multiprocessing.sharedctypes import Value


def get_datasets(self, datasetId=None, name=None, email=None, workspaceId=None):
    """Queries the workspace datasets based off provided parameters. Checks on datasetId, name, owner in this respective order within the specified workspace.
    If only workspace ID is provided, this will return all the datasets in a workspace. 
    
    Parameters
    ----------
    datasetId : str
        Dataset ID to filter.
    name : str 
        Dataset name.   
    email: str
        Owner of the dataset.
    workspaceId : str
        Workspace ID of the dataset's workspace. If none is provided, the current workspace will get used. 
    
    Returns
    -------
    str
        Information about the dataset based off the query parameters provided or a failure message. 
    """
    if self.check_logout(): return
    if datasetId is None: datasetId = ''
    if name is None: name = ''
    if email is None: owner = ''
    if workspaceId is None: workspaceId = self.workspace
    return self.ana_api.getDatasets(workspaceId=workspaceId, datasetId=datasetId, name=name, email=email)


def create_dataset(self, name, graphId, description='', runs=1, priority=1, seed=1, workspaceId=None):
    """Create a new dataset based off an existing staged graph. This will start a new job.
    
    Parameters
    ----------
    name: str
        Name for dataset. 
    graphId : str
        ID of the staged graph to create dataset from.
    description : str 
        Description for new dataset.
    runs : int
        Number of times a channel will run within a single job. This is also how many different images will get created within the dataset. 
    priority : int
        Job priority.
    seed : int
        Seed number.
    workspaceId : str
        Workspace ID of the staged graph's workspace. If none is provided, the current workspace will get used. 
    
    Returns
    -------
    str
        Success or failure message about dataset creation.
    """
    if self.check_logout(): return
    if name is None or graphId is None: return
    if description is None: description = ''
    if workspaceId is None: workspaceId = self.workspace
    return self.ana_api.createDataset(workspaceId=workspaceId, graphId=graphId, name=name, description=description, runs=runs, seed=seed, priority=priority)


def edit_dataset(self, datasetId, description=None, name=None, workspaceId=None):
    """Update dataset description.
    
    Parameters
    ----------
    datasetId : str
        Dataset ID to update description for.
    description : str 
        New description.
    name: str
        New name for dataset.
    workspaceId : str
        Workspace ID of the dataset to get updated. If none is provided, the current workspace will get used. 
    
    Returns
    -------
    str
        Success or failure message about dataset update.
    """
    if self.check_logout(): return
    if datasetId is None: return
    if workspaceId is None: workspaceId = self.workspace
    if name is None and description is None: return
    return self.ana_api.editDataset(workspaceId=workspaceId, datasetId=datasetId, name=name, description=description)
    

def delete_dataset(self, datasetId, workspaceId=None):
    """Delete an existing dataset.
    
    Parameters
    ----------
    datasetId : str
        Dataset ID of dataset to delete.
    workspaceId : str
        Workspace ID that the dataset is in. If none is provided, the current workspace will get used. 
    
    Returns
    -------
    str
        Success or failure message about dataset deletion.
    """
    if self.check_logout(): return
    if datasetId is None: return
    if workspaceId is None: workspaceId = self.workspace
    return self.ana_api.deleteDataset(workspaceId=workspaceId, datasetId=datasetId)
    

def download_dataset(self, datasetId, workspaceId=None, localDir=None):
    """Download a dataset.
    
    Parameters
    ----------
    datasetId : str
        Dataset ID of dataset to download.
    workspaceId : str
        Workspace ID that the dataset is in. If none is provided, the default workspace will get used. 
    
    Returns
    -------
    str
        Success or failure message about dataset download.
    """
    import requests, os
    if self.check_logout(): return
    if datasetId is None: datasetId
    if workspaceId is None: workspaceId = self.workspace
    if localDir is None: localDir = os.getcwd()
    elif not os.path.exists(localDir): raise Exception(f"Could not find directory {localDir}.")
    url = self.ana_api.downloadDataset(workspaceId=workspaceId, datasetId=datasetId)        
    fname = url.split('?')[0].split('/')[-1]
    with requests.get(url, stream=True) as downloadresponse:
        downloadresponse.raise_for_status()
        with open(os.path.join(localDir,fname), 'wb') as f:
            for chunk in downloadresponse.iter_content(chunk_size=8192): f.write(chunk)
    return os.path.join(localDir,fname)


def cancel_dataset(self, datasetId, workspaceId=None):
    """Stop a running job.
    
    Parameters
    ----------
    datasetId : str
        Dataset ID of the running job to stop.
    workspaceId: str
        Workspace ID of the running job. If none is provided, the default workspace will get used. 
    
    Returns
    -------
    str
        Success or error message about stopping the job execution.
    """
    if self.check_logout(): return
    if workspaceId is None: workspaceId = self.workspace
    return self.ana_api.cancelDataset(workspaceId=workspaceId, datasetId=datasetId)


def upload_dataset(self, filename, description=None, workspaceId=None):
    """Uploads user dataset. 
    
    Parameters
    ----------
    filename: str
        Path to the dataset folder or file for uploading. Must be zip or tar file types.
    workspaceId : str
        WorkspaceId to upload dataset to. Defaults to current.
    description : str
        Description for new dataset.
    
    Returns
    -------
    str
        success
    """
    import requests
    import os
    if self.check_logout(): return
    if filename is None: raise ValueError("Filename must be defined.")
    if description is None: description = ''
    if workspaceId is None: workspaceId = self.workspace
    if os.path.splitext(filename)[1] not in ['.zip', '.tar', '.gz']: raise Exception('Dataset Upload is only supported for zip, tar and tar.gz files.')
    if not os.path.exists(filename): raise Exception(f'Could not find file: {filename}') 
    fileinfo = self.ana_api.datasetUpload( workspaceId=workspaceId, filename=filename, description=description)
    datasetId = fileinfo['key'].split('/')[-1].split('.')[0]
    with open(filename, 'rb') as filebytes:
        files = {'file': filebytes}
        data = {
            "key":                  fileinfo['fields']['key'],
            "bucket":               fileinfo['fields']['bucket'],
            "X-Amz-Algorithm":      fileinfo['fields']['algorithm'],
            "X-Amz-Credential":     fileinfo['fields']['credential'],
            "X-Amz-Date":           fileinfo['fields']['date'],
            "X-Amz-Security-Token": fileinfo['fields']['token'],
            "Policy":               fileinfo['fields']['policy'],
            "X-Amz-Signature":      fileinfo['fields']['signature'],
        }
        response = requests.post(fileinfo['url'], data=data, files=files)
        if response.status_code != 204: 
            print(f'Failed to upload dataset, with status code: {response.status_code}')
            return False
    return datasetId


def get_dataset_runs(self, datasetId, state=None, workspaceId=None):
    """Shows all dataset run information to the user. Can filter by state.
    
    Parameters
    ----------
    datasetId: str
        The dataset to retrieve logs for.
    state: str
        Filter run list by status.
    workspaceId : str
        The workspace the dataset is in.
    
    Returns
    -------
    list[dict]
        List of run associated with datasetId.
    """
    if self.check_logout(): return
    if workspaceId is None: workspaceId = self.workspace
    runs = self.ana_api.getDatasetRuns(workspaceId=workspaceId, datasetId=datasetId, state=state)
    if runs: return runs
    else: return None
                

def get_dataset_log(self, datasetId, runId, saveLogFile=False, workspaceId=None):
    """Shows dataset log information to the user.
    
    Parameters
    ----------
    datasetId: str
        The dataset the run belongs to.
    runId: str
        The run to retrieve the log for.
    saveLogFile: bool
        If True, saves log file to current working directory.
    workspaceId: str
        The workspace the run belongs to.
    
    Returns
    -------
    list[dict]
        Get log information by runId
    """
    import json 

    if self.check_logout(): return
    if workspaceId is None: workspaceId = self.workspace
    log = self.ana_api.getDatasetLog(workspaceId=workspaceId, datasetId=datasetId, runId=runId)
    if log['log']: 
        if saveLogFile:
            with open (f"{datasetId}-{log['run']}.log",'w+') as logfile:
                for line in json.loads(log['log']): logfile.write(f"{line['message']}\n")
            print(f"Saved log to {datasetId}-{log['run']}.log")
        else: return log['log']
    else: return None